/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* ai.c
   Contains code to manage computer players */

#include <stdio.h>
#include <math.h>
#include <string.h>

#include "ai.h"
#include "world.h"
#include "game.h"
#include "round.h"
#include "turn.h"
#include "player.h"
#include "vector.h"
#include "tools.h"
#include "map.h"
#include "text.h"
#include "pathfind.h"
#include "player_stats.h"

static float target_enemy(Player *player);
static void update_ai_fire(World *world, Player *player, float dt);
static void update_ai_move(World *world, Player *player, float dt);
static void search_new_destination(World *world, Player *player);
static void buy_all_weapons(World *world, Player *player);
static char *select_a_name(World *world);
static Tank *select_a_tank(World *world);

static void assess_situation(Player *player);
static void update_temp_stats(Player *player);
static void follow_path(Player *player, float dt);
static void attack_enemy(Player *player, float dt);

/* Time (in seconds) of what range of time the AI should take per turn */
#define MIN_AI_WAIT 10.0
#define MAX_AI_WAIT 30.0

/* List of AI names to choose from */
char *ai_names[] =
{"Battlin'Joe", "Killer", "Peace & Love", "Hexasoft", "Neutronium",
 "Extropy", "Anihilator", "Coyote", "Big Bertha", "Lucie", "H.A.L.",
 "Runner", "Grrrrr!", "GoodBye", "Speedy", "Alien", "Arkonen",
 "Felix", "Sergei", "Tornado", "Taz", "H2O", "-=X|X=-",
};
int nb_ai_names = sizeof(ai_names) / sizeof(char*);

/* Make a new AI */
AI *make_ai(Player *player)
{
  AI *ai;

  if(!player || player->ai)
    return NULL;

  ai = malloc(sizeof(AI));
  if(!ai)
  {
    Neprintf("Error: Unable to allocate space for AI\n");
    return NULL;
  }

  /* Init the values */
  ai->player = player;
  ai->turn = 0;
  ai->turn_time = 0.0;
  ai->situation_assessed = FALSE;
  ai->path = NULL;
  ai->target = NULL;

  ai->dest[0] = 0.;
  ai->dest[1] = 0.;
  ai->dest[2] = 0.;
  ai->move  = AIM_WAITING;
  ai->fight = AIF_WAITING;
  ai->round_state = AIR_WAITING;
  ai->incoming = FALSE;

  player->ai = ai;

  return ai;
}

/* Return TRUE if the player is AI controlled */
int is_ai_controlled(Player *player)
{
  if(!player)
    return FALSE;

  if(player->player_type != PLAYER_HUMAN &&
     player->player_type != PLAYER_HUMAN_NET &&
     player->ai)
    return TRUE;

  return FALSE;
}

/* Update the players controlled by an AI */
void update_ai_players(World *world, float dt)
{
  Player *player = world->player_list;

  while(player)
  {
    /* If player is controlled by an AI... */
    if(is_ai_controlled(player))
    {
      /* Update player movement */
      update_ai_move(world, player, dt);

      /* Update player firing */
      update_ai_fire(world, player, dt);
    }

    player = player->next;
  }
}

/* Get an AI to target its enemy */
static float target_enemy(Player *player)
{
  Player *target;
  float *barrel_dir, enemy_dir[3], dir_diff[3];
  float precision, adir;
  float range, v[3];

  if(!player || !player->ai || !player->ai->target)
    return(0.);

  target = player->ai->target;

  /* Get the direction the barrel is pointing in */
  barrel_dir = glassGetActivePointDir(player->tank->model, player->tank->barrel);

  /* Get a unit direction vector to the target */
  _vector_sub(player->pos, target->pos, enemy_dir);
  normalize_vector(enemy_dir);

  /* Get the difference in these two vectors --
     how much they are pointing in the same direction */
  _vector_add(barrel_dir, enemy_dir, dir_diff);

  /* Remove the vertical component -- this only affects range */
  dir_diff[1] = 0.0;

  /* If this vector is small enough, then tank roughly pointing in the correct direction */
  switch (player->player_type)
    {/* select a precision regards to AI skill */
    case PLAYER_AI_STUPID:
      precision = 0.4;
      break;
    case PLAYER_AI_MEDIUM:
      precision = 0.3;
      break;
    case PLAYER_AI_GOOD:
      precision = 0.2;
      break;
    case PLAYER_AI_GOD:
      precision = 0.1;
      break;
    default:/* may not occur */
      precision = 0.2;
      break;
    }
  if(vector_norm(dir_diff) < precision*precision) /* using the squared norm (faster) */
  {
    /* Get the force right to hit there */
    /* TEMP: Use a ballpark figure */
    range = vector_2norm(_vector_sub(player->pos, target->pos, v));
    player->muzzle_vel = sqrt(-4.0 / (player->game->gravity * range));
    player->thrust = 0.5 * range;

    /* Fire the tank */
    player_end_turn(player);
    player->ai->situation_assessed = FALSE;
    return(0.0);
  }
  /* not the good direction: compute the angle to turn */
  enemy_dir[0] *= -1;
  enemy_dir[1] *= -1; /* vector from tank to enemy */
  enemy_dir[2] *= -1;
  /* wrong side */
  if (enemy_dir[0]*barrel_dir[0] + enemy_dir[2]*barrel_dir[2] <= 0.)
    {
    return(20.); /* bad side: turn fast in any direction */
    }
  else
    {/* turn right or left ? */
    adir = -enemy_dir[2]*barrel_dir[0] + enemy_dir[0]*barrel_dir[2];
    if (adir <= 0.)
      {/* right */
      if (adir < 0.1)
        return(-15.0); /* far: turn fast */
      else
        return(-7.0); /* near: turn slower */
      }
    else
      {/* left */
      if (adir > 0.1)
        return(15.0); /* far: turn fast */
      else
        return(7.0); /* near: turn slower */
      }
    }
  /* never occur, just for prevent warnings */
  return(0.);
}

/* Update AI player firing */
static void update_ai_fire(World *world, Player *player, float dt)
{
  float fdir[3], temp[3], turning;
  Player *opp;
  int i;

  if(!player || !player->ai || !player->game)
    return;

  /* Check player is an alive AI player */
  if((player->player_type == PLAYER_HUMAN)||
     (player->player_type == PLAYER_HUMAN_NET))
    return;
  if(player->player_state == PLAYER_DEAD)
    return;

  /* Increase flying weapon duration */
  if (player->ai->last_fire >= 0.0)
    player->ai->last_fire += dt;

  /* If we have fired, dont treat that again before some times */
  if (player->ai->last_fire >= 0.)
    {
    if (player->ai->last_fire < 8.)
      return;
    else
      player->ai->last_fire = -1.;
    }

  /* If not on the ground, dont fire */
  if(player->tank_state != ON_GROUND)
    return;

  /* Always turn the turret */
  /* inc_player_turret_heading(player, 16.0f*dt); */

  /* Get the fire direction */
  /*get_fire_direction(player, fdir);*/
  turning = target_enemy(player);
  /* I changed this: target_enemy returns the amount of rotation to
       apply to turret in order to reduce angle distance. 0. means
       that target is reached */
  if (turning != 0.)
    inc_player_turret_heading(player, turning*dt);
  return;

  /* if a player != than us is in the direction... */
  opp = player->world->player_list;
  while(opp != NULL)
    {/* dont fire on use and on deads */
    if ((opp != player)&&(opp->player_state != PLAYER_DEAD))
      {
      _vector_sub(opp->pos, player->pos, temp);
      /* too near to fire! */
      if (vector_2norm(temp) < 5.)
        {
        return;
        }
      normalize_vector(temp);
      /* roughly the good direction */
      if ((nABS(-fdir[2]*temp[0] + fdir[0]*temp[2]) < 0.08)&& /* in the line */
          (fdir[0]*temp[0] + fdir[2]*temp[2] > 0.))  /* and the same side */
        {
        /* ok. we will not fire all time, so let randomize it */
	  if (random_float(0.,1.) < .1) /* 10% prob. to fire */
          {
          /* select a weapon */
          i = random_integer(0, 50);
          for(; i>=0; i--)
            {
            cycle_weapon_fowards(player);
            }
          /* now just fire */
          player_end_turn(player);
    player->ai->situation_assessed = FALSE;
          /* start time counter */
          player->ai->last_fire = 0.01;
          return;
          }
        }
      }
    opp = opp->next;
    }
}

/* Update AI players movement */
static void update_ai_move(World *world, Player *player, float dt)
{
  AI *state = player->ai;
  float temp[3], ff, aspeed;

  if(!player)
    return;

  /* Player is dead */
  if(player->player_state == PLAYER_DEAD)
    return;

  /* human control */
  if((player->player_type == PLAYER_HUMAN)||
     (player->player_type == PLAYER_HUMAN_NET))
    return;

  /* Beginning of the AI decision system */
  
  /* if not moving, select a destination and run */
  if (state->move == AIM_WAITING)
    {
    /* search a destination on the map, not under sea level
       and not too near from an other tank */
    search_new_destination(world, player);
    /* switch to accelerating mode */
    state->move = AIM_ACCELERATING;
    return;
    }

  /* moving: compute distance to target to see allowed speed */
  _vector_sub(state->dest, player->pos, temp);
  ff = vector_2norm(temp);
  if (ff > 100.)
    aspeed = .5;
  else if (ff < 10.)
    aspeed = .05;
  else if (ff < 10.)
    aspeed = .05;
  else if (ff < 20.)
    aspeed = .075;
  else if (ff < 40.)
    aspeed = .1;
  else if (ff < 60.)
    aspeed = .15;
  else if (ff < 80.)
    aspeed = .2;
  else
    aspeed = .3;

  /* if destination is under water (after explosion), choose an other */
  if(get_height(player->game->map, state->dest[0], state->dest[2]) <= player->game->sea_level)
    {
    state->move = AIM_WAITING;
    return;
    }

  /* if moving, turns to reach the destination */
  if (state->move != AIM_WAITING)
    {
    _vector_sub(state->dest, player->pos, temp);
    /* wrong side */
    if (temp[0]*player->dir[0] + temp[2]*player->dir[2] <= 0.)
      {
      inc_player_heading(player, 2.0f);
      player->speed = player->tank->speed * 0.05; /* break! */
      }
    else
      {/* turn right or left ? */
      if (-temp[2]*player->dir[0] + temp[0]*player->dir[2] <= 0.)
        {/* right */
        inc_player_heading(player, -20.0f*dt);
        }
      else
        {/* left */
        inc_player_heading(player, 20.0f*dt);
        }
      }
    }

  /* if we are trying to reach a position, continue */
  if (state->move == AIM_ACCELERATING)
    {
    inc_player_accel(player, 1.5f*dt);
    if (player->speed >= player->tank->speed * aspeed)
      {/* current speed reached */
      state->move = AIM_MOVINGTO;
      }
    return;
    }

  /* need to slowdown ? */
  if (state->move == AIM_MOVINGTO)
    {
    if (player->speed >= player->tank->speed * aspeed)
      {
      inc_player_speed(player, -1.0f*dt);
      }
    }

  /* if we reach the destination, break */
  if (state->move == AIM_MOVINGTO)
    {
    _vector_sub(state->dest, player->pos, temp);
    ff = vector_2norm(temp);
    if (ff < 10.)
      {
      set_player_speed(player, 0.0f);
      set_player_accel(player, 0.0f);
      state->move = AIM_BREAKING;
      }
    return;
    }

  /* if breaking and speed = 0, waiting mode */
  if (state->move == AIM_BREAKING)
    {
    if (vector_2norm(player->vel) < 1.5)
      {
      state->move = AIM_WAITING;
      }
    return;
    }
}


/* search a 'safe' destination to moves */
static void search_new_destination(World *world, Player *player)
{
  int x[16], y[16];
  float dists[16], dst, max;
  int i, sel;
  Player *pls;
  Map *map = player->game->map;

  /* try 16 destinations and select the best one */
  for(i=0; i<16; i++)
    {
    do
      {
      x[i] = random_integer(0, map->size_x - 1);
      y[i] = random_integer(0, map->size_y - 1);
      }
    while(map->pos[x[i]][y[i]][1] < player->game->sea_level);
    }
  /* compute distances to other players */
  for(i=0; i<16; i++)
    {
    pls = world->player_list;
    dists[i] = 9999999.;
    while(pls != NULL)
      {
      if (pls != player)
        {
        dst = (x[i]*map->x_resolution-pls->pos[0])*(x[i]*map->x_resolution-pls->pos[0]) +
              (y[i]*map->y_resolution-pls->pos[0])*(y[i]*map->y_resolution-pls->pos[2]);
        if (dst < dists[i])
          dists[i] = dst;
        }
      pls = pls->next;
      }
    }
  /* select the one which is far away from other */
  max = 0.;
  sel = 0;
  for(i=1; i<16; i++)
    {
    if (dists[i] > max)
      {
      max = dists[i];
      sel = i;
      }
    }
  /* set the destination */
  player->ai->dest[0] = x[sel]*map->x_resolution;
  player->ai->dest[1] = 0.; /* not used */
  player->ai->dest[2] = y[sel]*map->y_resolution;
}

/* Configure the AI players */
void player_config_ai(World *world)
{
  Player *player = world->player_list;

  /* For each AI player */
  while(player)
    {
    if ((player->player_type != PLAYER_HUMAN)&&
        (player->player_type != PLAYER_HUMAN_NET))
      {
      /* select a name */
      set_player_name(player, select_a_name(world));
      /* select a tank */
      player->tank = select_a_tank(world);
      Nprintf("  AI player '%s' choose tank '%s'.\n",
              player->name, player->tank->name);
      }
    player = player->next;
    }
}

/* Get an AI player to select its name, tank etc */
void config_ai_player(Player *player)
{
  Tank *tank;

  if(!player || !player->world)
    return;

  set_player_name(player, select_a_name(player->world));
  tank = select_a_tank(player->world);
  if(tank)
    set_player_tank(player, tank->name);
  else
    set_player_tank(player, NULL);
}

/* Get an AI to buy weapons */
void ai_buy_weapons(Player *player)
{
  if(!player || !player->world)
    return;

  buy_all_weapons(player->world, player);
}

/* Get an AI to buy items */
void ai_buy_items(Player *player)
{
  if(!player || !player->world)
    return;

  Nprintf("AI buying items unsupported\n");
}

/* configure AI weapons */
void player_shop_ai(World *world)
{
  Player *player = world->player_list;

  /* for each AI player */
  while(player != NULL)
    {
    if ((player->player_type != PLAYER_HUMAN)&&
        (player->player_type != PLAYER_HUMAN_NET))
      {
	buy_all_weapons(world, player);
      }
    player = player->next;
    }
}


/* buy some weapons until no more money */
static void buy_all_weapons(World *world, Player *player)
{
  int i, old_cash, again;
  Weapon *weapon;

  Nprintf("%s buys:\n", player->name);
  do
    {
    old_cash = player->cash;
    /* check if we can still buy something */
    weapon = player->weapon_list;
    again = FALSE;
    while(weapon != NULL)
      {
      /* still a weapon that we can buy (dont count 'No Weapon') */
      if ((weapon->munition->cost <= player->cash)&&(weapon->munition->cost > 0))
        {
        again = TRUE;
        break;
        }
      weapon = weapon->next;
      }
    if (!again)
      break;
    /* select a weapon with a price < than our chash */
    do
      {
      weapon = player->weapon_list;
      i = random_integer(0, 50);
      for(; i>=0; i--)
        {
        weapon = weapon->next;
        if (weapon == NULL)
          weapon = player->weapon_list;
        }
      }
    while((weapon->munition->cost > player->cash)||(weapon->munition->cost <= 0));
    player->selected_weapon = weapon;
    buy_weapon(player, weapon->munition->name, 1);
    Nprintf("  %s ($%d remaining)\n", weapon->munition->name, player->cash);
    }
  while(player->cash != old_cash);
}


/* Choose a name for an AI player */
static char *select_a_name(World *world)
{
  Player *player;
  int i = 0, ok;

  do
  {
    i = ((float)rand() / RAND_MAX) * (nb_ai_names - 1);
    /* Check if the name is not used */
    ok = TRUE;
    player = world->player_list;
    while(player)
    {
      if(player->name && !strcmp(player->name, ai_names[i]))
      {
        ok = FALSE;
        break;
      }

      player = player->next;
    }
  } while(!ok);

  return(ai_names[i]);
}

/* Randomly select a tank */
static Tank *select_a_tank(World *world)
{
  Tank *tank = world->tank_list;
  int i;

  i = random_integer(0, 50);
  for(; i >= 0; i--)
  {
    tank = tank->next;
    if(!tank)
      tank = world->tank_list;
  }
  return(tank);
}

/** round-based AI controls **/

/* Update AI player for round-based game */
void update_ai_player_round(Player *player, float dt)
{
  World *world;
  Game *game;
  AI *ai;

  if(!player || !player->world || !player->ai)
    return;

  world = player->world;
  game = player->game;
  ai = player->ai;

  /* If a new round has started, must reasses the situation */
  if(ai->turn != game->round->turn)
  {
    ai->turn = game->round->turn;
    ai->turn_time = 0.0;
    ai->situation_assessed = FALSE;
  }

  ai->turn_time += dt;

  /* If situation hasn't been assessed, do that
     i.e. See what things are like this round,
     and decide what to do this round */
  if(!ai->situation_assessed)
    assess_situation(player);
  if(!ai->situation_assessed)
    return;

  /* If currently moving along a path */
  if(ai->path)
  {
    follow_path(player, dt);
    return;
  }

  /* If targeting an enemy */
  if(ai->target)
  {
    attack_enemy(player, dt);

    /* Don't take longer that MAX_AI_WAIT seconds */
    if(ai->turn_time < MAX_AI_WAIT)
      return;
  }

  /* Always take at least MIN_AI_WAIT seconds */
  if(ai->turn_time < MIN_AI_WAIT)
    return;

  /* AI doesn't want to do anything */
  Nprintf("Player %s skips turn\n", player->name);
  player_end_turn(player);
}

/* Get an AI player to assess the situation it is in, and decide
   what to do next */
static void assess_situation(Player *player)
{
  Game *game;
  Map *map;
  AI *ai;
  Player *target;
  PlayerStat *stat;
  Tank *tank;
  float easiest = 0.0;
  pathnode *path;
  float x, y, drive_time, t, range;
  int count, has_engine;

  if(!player || !player->ai || !player->world)
    return;

  game = player->game;
  map = game->map;
  ai = player->ai;
  tank = player->tank;
  has_engine = tank->has_engine;

  /* 1. If falling, do nothing */
  if(player->tank_state == IN_AIR ||
     player->tank_state == PARACHUTE ||
     player->tank_state == UNDERWATER)
    return;

  Nprintf("%s assess the situation...\n", player->name);

  /* Assume that this will decide on a solution */
  ai->situation_assessed = TRUE;

  /* Assess the other players at this moment */
  update_temp_stats(player);

  /* Remove an existing path */
  free_path(ai->path);
  ai->path = NULL;

  /* Stop attacking an enemy */
  ai->target = NULL;

  /* Decide what to do */

  /* 1. If falling... (see above) */

  /* 2. If underwater, head for dry ground */
  if(player->tank_state == ON_SEAFLOOR && has_engine)
  {
    Nprintf("player %s decides to head for dry ground\n", player->name);
    ai->path = find_dry_path(player);
    ai->situation_assessed = TRUE;
    return;
  }

  /* 3. Shoot the easiest target */
  /* Find the easiest target */
  stat = player->player_stats;
  target = NULL;
  while(stat)
  {
    if(stat->player != player && stat->easiness > easiest)
    {
      easiest = stat->easiness;
      target = stat->player;
    }

    stat = stat->next;
  }

  /*printf("Easiest target = %f\n", easiest);*/
  /* If found a target and it is easy enough, attack them */
  if(target && easiest > 1.0)
  {
    ai->target = target;
    Nprintf("player %s decides to fire at %s\n", player->name, target->name);
    ai->situation_assessed = TRUE;
    return;
  }

  /* 4. If all else fails, move to a better shot */

  /* No engine, so just skip the turn */
  if(!has_engine)
    return;

  /* Find a destination */

  /* How long they have to drive in one turn */
  drive_time = game->driving_time;

  /* Approx. how far they can get in one turn */
  /* Time taken to get to max speed */
  t = tank->speed / tank->acc;
  /* If wont get to max speed*/
  if(t > drive_time)
    range = 0.5 * tank->acc * drive_time*drive_time;
  else /* Will hit max speed */
    range = 0.5 * tank->acc * t*t + tank->speed * (drive_time - t);

  /* Try ten above water points to move to, otherwise give up */
  count = 0;
  do
  {
    do
    {
      x = player->pos[0] + random_float(0.0, range) - 0.5 * range;
      y = player->pos[2] + random_float(0.0, range) - 0.5 * range;
    } while(!in_bounds(map, x, y) || get_height(map, x, y) <= game->sea_level);

    /* Find a dry path to this destination */
    path = find_path(player, x, y);

    count++;
  } while(!path && count < 10);

  if(path)
  {
    ai->path = path;
    Nprintf("Player %s decides to move to (%.3f, %.3f)\n", player->name, x, y);
  }

  ai->situation_assessed = TRUE;
}

/* Update the temporary statistics on the opposing players */
static void update_temp_stats(Player *player)
{
  Player *p;
  PlayerStat *stat;
  Game *game;
  Map *map;
  float *pos1, v[3], range = 0.0, threat, easiness;
  float min_health = 1000000.0, max_health = 0.0, min_range = 1000000000000.0;

  if(!player || !player->game || !player->game->map)
    return;

  game = player->game;
  map = game->map;
  pos1 = player->pos;

  /* Go through once to get the constant facts and find the (health etc) limits */
  stat = player->player_stats;
  while(stat)
  {
    p = stat->player;

    /* If alive */
    if(p->player_state)
    {
      /* Get the range */
      _vector_sub(pos1, p->pos, v);
      range = vector_2norm(v);
      stat->range = range;

      /* Find the closest */
      if(p != player && range < min_range)
	min_range = range;

      /* Check if player is visible */
      if(get_map_intersection_long(map, pos1, p->pos, -1))
	stat->visible = FALSE;
      else
	stat->visible = TRUE;

      /* Find the strongest/weakest etc */
      if(p->health > max_health)
	max_health = p->health;
      if(p->health < min_health)
	min_health = p->health;
    }

    stat = stat->next;
  }

  /* Go through a second time and toggle the strongest/weakest etc flags */
  stat = player->player_stats;
  while(stat)
  {
    p = stat->player;

    if(p != player && p->player_state)
    {
      if(p->health == min_health)
	stat->weakest = TRUE;
      else
	stat->weakest = FALSE;

      if(p->health == max_health)
	stat->strongest = TRUE;
      else
	stat->strongest = FALSE;

      if(stat->range == min_range)
	stat->closest = TRUE;
      else
	stat->closest = FALSE;
    }

    stat = stat->next;
  }

  /* Go through a third time and assess the threat/easiness of the players */
  stat = player->player_stats;
  while(stat)
  {
    p = stat->player;
    threat = 0.0;
    easiness = 0.0;

    /* If alive */
    if(p != player && p->player_state)
    {
      /* NOTE: All these constants will define the AI */
      threat =
	0.01 * stat->damage_taken +
	0.5 * stat->losses +
	0.01 * stat->range;
      if(stat->closest)
	threat += 0.1;
      if(stat->strongest)
	threat += 0.2;

      easiness =
	0.01 * stat->damage_inflicted + /* If have done 100 damage or more */
	0.5 * stat->kills +             /* Or killed them twice */
        150.0 / stat->range;            /* Or within 150m */
      if(stat->weakest)                 /* Get the weakling! */
	easiness += 0.2;
    }

    stat->threat = threat;
    stat->easiness = easiness;

    stat = stat->next;
  }
}

/* Have a player follow the path they have */
static void follow_path(Player *player, float dt)
{
  AI *ai;
  pathnode *node;
  /* The map co-ordinates of the player and waypoint */
  float px, py, wx, wy;
  /* The difference between the two points */
  float dx, dy;
  /* The direction the tank is pointing, and the direction it needs to point */
  float *dir, dirx, diry, tdirx, tdiry;
  float len, sin_theta;
  float *vel, speed;
  float target_speed = 100.0;

  if(!player || !player->ai || !player->ai->path)
    return;

  ai = player->ai;

  node = ai->path;
  px = player->pos[0];
  py = player->pos[2];
  wx = node->x;
  wy = node->y;
  dx = wx - px;
  dy = wy - py;

  start_player_engine(player);

  /* Remove the current waypoint if already there */
  if((dx*dx + dy*dy) < 3.0)
  {
    ai->path = pop_node(ai->path);
    /*set_player_brakes(player, TRUE);*/
    /* If finished move, player is done */
    if(!ai->path)
      player_end_turn(player);
    return;
  }

  /* Point at the waypoint */
  /* Direction player is pointing now */
  dir = player->dir;
  len = 1.0 - dir[1]*dir[1];
  dirx = dir[0] / len;
  diry = dir[2] / len;

  /* Unit direction to target */
  len = sqrt(dx*dx + dy*dy);
  tdirx = dx / len;
  tdiry = dy / len;

  /* Take cross product */
  sin_theta = dirx*tdiry - diry*tdirx;

  /* If pointing the other way... */
  if(dirx*tdirx + diry*tdiry < 0.0)
  {
    if(sin_theta > 0.0)
      sin_theta = 1.0;
    else
      sin_theta = -1.0;
  }

  /* If not pointing in correct direction, rotate */
  if(!(fabs(sin_theta) < 0.001))
  {
    /*set_player_brakes(player, TRUE);*/
    inc_player_heading(player, 50.0 * -sin_theta * dt);
    target_speed = 5.0 * (1.0 - fabs(sin_theta));
  }

  /* The current speed of the tank */
  vel = player->vel;
  speed = sqrt(vel[0]*vel[0]*dir[0]*dir[0] + vel[2]*vel[2]*dir[2]*dir[2]);

  /* Move to the current waypoint */
  set_player_accel(player, 0.5 * (target_speed - speed));
}

/* Attack the current target */
static void attack_enemy(Player *player, float dt)
{
  AI *ai;
  Player *target;
  float dir[3], dirx, diry;
  float *tpos, *pos, dx, dy, len;
  float tdirx, tdiry;
  float sin_theta;
  int i;

  if(!player || !player->ai || !player->ai->target)
    return;

  ai = player->ai;
  target = ai->target;

  /* Point at the target */

  /* Choose a random pitch */
  inc_player_turret_pitch(player, random_float(-0.1, 0.1));

  /* Position of player and target */
  pos = player->pos;
  tpos = target->pos;
  dx = tpos[0] - pos[0];
  dy = tpos[2] - pos[2];

  /* Direction player is pointing now */
  get_fire_direction(player, dir);
  len = 1.0 - dir[1]*dir[1];
  dirx = dir[0] / len;
  diry = dir[2] / len;

  /* Unit direction to target */
  len = sqrt(dx*dx + dy*dy);
  tdirx = dx / len;
  tdiry = dy / len;

  /* Take cross product */
  sin_theta = dirx*tdiry - diry*tdirx;

  /* If pointing the other way... */
  if(dirx*tdirx + diry*tdiry < 0.0)
  {
    if(sin_theta > 0.0)
      sin_theta = 1.0;
    else
      sin_theta = -1.0;
  }

  /* If not pointing in correct direction, rotate turret */
  /* TODO: This should be based on the AI's accuracy */
  if(!(fabs(sin_theta) < 0.01))
  {
    inc_player_turret_heading(player, 50.0 * -sin_theta * dt);
    return;
  }

  /* Always take at leasy MIN_AI_WAIT seconds */
  if(ai->turn_time < MIN_AI_WAIT)
    return;

  /* Select a random weapon */      
  i = random_integer(1, 50);
  for(; i >= 0; i--)
    cycle_weapon_fowards(player);

  /* Fire! */
  player_end_turn(player);
}

/* Draw what the AI is thinking */
void draw_ai(AI *ai)
{
  if(!ai)
    return;

  /*if(world->draw_ai)*/
    draw_path(ai->player->world, ai->path);
}

/* Free an AI */
void free_ai(AI *ai)
{
  Nfree(ai);
}
