/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* bird.c
   Contains code to manage birds in the map */

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include "bird.h"
#include "world.h"
#include "game.h"
#include "vector.h"
#include "text.h"
#include "tools.h"
#include "update.h"
#include "map.h"

/* pre-computed cos/sin */
#define PC_NB 32
static float pc_cos[PC_NB+1] = {0.};
static float pc_sin[PC_NB+1] = {0.};
static float pc_angle[PC_NB+1] = {0.};
static float pc_step=2.*3.14159 / PC_NB;

static Bird *allocate_bird(Round *round, int type);
static void free_bird(Bird *bird);

/* init the bird system */
int init_birds(World *world)
{
  int i;
  float angle = 0.;

  for(i=0; i<PC_NB+1; i++)
    {
    pc_cos[i] = (float)cos((float)angle);
    pc_sin[i] = (float)sin((float)angle);
    pc_angle[i] = angle*360./(2.*3.14159);
    angle += pc_step;
    }

  return(TRUE);
}

/* free the bird system */
void free_birds(World *world)
{
}


/* update existing birds */
void update_birds(Round *round, float dt)
{
  Bird *bird = round->bird_list;

  /* birds system disabled */
  if(!round->world->is_bird)
    return;

  while(bird)
  {
    update_bird(bird, dt);
    bird = bird->next;
  }
}
/* update a bird */
void update_bird(Bird *bird, float dt)
{
  Map *map;
  int x, y;
  float temp[3];
  float sea_level;

  map = bird->game->map;
  sea_level = bird->game->sea_level;

  if (bird->type == BIRD_LINE)
    {
    /* update position */
    bird->pos[0] += bird->dir[0]*bird->speed*dt;
    bird->pos[2] += bird->dir[2]*bird->speed*dt;
    bird->pos[1] = nMAX(get_height(map, bird->pos[0], bird->pos[2]) + 10.,
                	sea_level + 10.);

    /* target reached ? */
    if (vector_distance(bird->pos, bird->dest) < 8.)
      {/* select a new target */
      do
	{
	x = (int)(((float)rand() / RAND_MAX) * (map->size_x - 1));
	y = (int)(((float)rand() / RAND_MAX) * (map->size_y - 1));
	}
      while (map->pos[x][y][1] > sea_level);
      bird->dest[0] = x * map->x_resolution;
      bird->dest[2] = y * map->y_resolution;
      bird->dest[1] = nMAX(map->pos[x][y][1] + 10., sea_level + 10.);
      /* switch to circle mode */
      /* bird->type = BIRD_CIRCLE; */
      }

    /* update bird direction */
    _vector_sub(bird->dest, bird->pos, temp);
    normalize_vector(temp);
    bird->dir[0] = (bird->dir[0] + dt*temp[0] ) / (1. + dt);
    bird->dir[1] = (bird->dir[1] + dt*temp[1] ) / (1. + dt);
    bird->dir[2] = (bird->dir[2] + dt*temp[2] ) / (1. + dt);
    normalize_vector(bird->dir);
    }
  else if (bird->type == BIRD_CIRCLE)
    {
    bird->pos[0] += bird->dir[0]*bird->speed*dt;
    bird->pos[2] += bird->dir[2]*bird->speed*dt;
    bird->pos[1] = nMAX(get_height(map, bird->pos[0], bird->pos[2]) + 10.,
                	sea_level + 10.);
    }
}

/* draw the birds */
void draw_birds(Round *round)
{
  Bird *bird = round->bird_list;

  /* birds system disabled */
  if(!round->world->is_bird)
    return;

  while(bird)
  {
    draw_bird(bird);
    bird = bird->next;
  }
}

/* draw a bird */
void draw_bird(Bird *bird)
{
  GLfloat bird_color[4]={1,1,1,1};
  float roty, dists[PC_NB], min, dir[3];
  float total_elapsed_time = bird->world->total_elapsed_time;
  int i, res;

  /* compute vertical rotation (approx.) */
  dir[2] = bird->dir[0];
  dir[0] = bird->dir[2];
  dir[1] = 0.;
  normalize_vector(dir);
  for(i=0; i<PC_NB; i++)
    {
    dists[i] = (dir[0] - pc_cos[i])*(dir[0] - pc_cos[i]) +
               (dir[2] - pc_sin[i])*(dir[2] - pc_sin[i]);
    }
  min = dists[0];
  res = 0;
  for(i=1; i<PC_NB; i++)
    {
    if (dists[i] < min)
      {
      res = i;
      min = dists[i];
      }
    }
  roty = pc_angle[res];
  bird->angle = roty;

  glColor4fv(bird_color);
  glPushAttrib(GL_ENABLE_BIT);
  glDisable(GL_TEXTURE_2D);
  glDisable(GL_CULL_FACE);
  glDisable(GL_BLEND);
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, bird_color);
  glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 128.0f);

  glPushMatrix();
  glTranslatef(bird->pos[0], bird->pos[1], bird->pos[2]);
  glRotatef(roty, 0, 1, 0);

  glScalef(.7, .7, .7);

  glBegin(GL_TRIANGLES);
  glNormal3f(0,1,0); /* to correct */
  glVertex3f(-2, 0.9*cos(bird->wspeed*total_elapsed_time), 0);
  glVertex3f(0, 0, .4);
  glVertex3f(0, 0,-.3);
  glNormal3f(0,1,0); /* to correct */
  glVertex3f(2, 0.9*cos(bird->wspeed*total_elapsed_time), 0);
  glVertex3f(0, 0, .4);
  glVertex3f(0, 0,-.3);
  glNormal3f(1,0,0); /* to correct */
  glVertex3f(-2, 0.9*cos(bird->wspeed*total_elapsed_time), 0);
  glVertex3f(0, 0, 0);
  glVertex3f(0, -.2, 0);
  glNormal3f(1,0,0); /* to correct */
  glVertex3f(2, 0.9*cos(bird->wspeed*total_elapsed_time), 0);
  glVertex3f(0, 0, 0);
  glVertex3f(0, -.2, 0);
  glEnd();

  glPopMatrix();
  glPopAttrib();
}

/* create the birds for the current map */
void create_birds(Round *round, int nb)
{
  int i;
  Bird *tmp = NULL;

  /* no bird */
   if (nb == 0)
     {
     round->bird_list = NULL;
     return;
     }

  Nprintf("  Creating %d birds...\n", nb);
  for(i=0; i<nb; i++)
    {
    tmp = allocate_bird(round, 0);
    if (tmp == NULL)
      {
      Neprintf("  Birds creation aborted.\n");
      round->bird_list = NULL;
      return;
      }
    tmp->next = round->bird_list;
    if (round->bird_list != NULL)
      round->bird_list->last = tmp;
    round->bird_list = tmp;
    }
  Nprintf("  Done.\n");
}

/* Remove all birds */
void clear_birds(Round *round)
{
  Bird *bird = round->bird_list;
  Bird *tmp;

  while(bird != NULL)
    {
    tmp = bird->next;
    free_bird(bird);
    bird = tmp;
    }
}

/* delete a bird */
void free_bird(Bird *bird)
{
  Nfree(bird);
}

/* allocate a bird */
Bird *allocate_bird(Round *round, int type)
{
  Map *map;
  int x, y;
  Bird *tmp = malloc(sizeof(Bird));
  float sea_level;

  map = round->map;
  sea_level = round->game->sea_level;

  if (tmp == NULL)
    {
    Neprintf("  Not enough memory to allocate a bird.\n");
    return(NULL);
    }

  tmp->world = round->game->world;
  tmp->game = round->game;
  tmp->round = round;

  tmp->next = tmp->last = NULL;
  tmp->type = type;
  tmp->speed = 5.;
  tmp->wspeed = 2.*3.14159/(4. + ((float)rand() / RAND_MAX) - .5);
  tmp->type  = BIRD_LINE;
  tmp->angle = 0.;

  /* search a position */
  do
    {
    x = (int)(((float)rand() / RAND_MAX) * (map->size_x - 1));
    y = (int)(((float)rand() / RAND_MAX) * (map->size_y - 1));
    }
  while (map->pos[x][y][1] > sea_level);
  tmp->pos[0] = x * map->x_resolution;
  tmp->pos[2] = y * map->y_resolution;
  tmp->pos[1] = nMAX(map->pos[x][y][1] + 10., sea_level + 10.);

  /* search a target */
  do
    {
    x = (int)(((float)rand() / RAND_MAX) * (map->size_x - 1));
    y = (int)(((float)rand() / RAND_MAX) * (map->size_y - 1));
    }
  while (map->pos[x][y][1] > sea_level);
  tmp->dest[0] = x * map->x_resolution;
  tmp->dest[2] = y * map->y_resolution;
  tmp->dest[1] = nMAX(map->pos[x][y][1] + 10., sea_level + 10.);

  /* Fly in a random direction */
  _make_random_vector(tmp->dir);

  return(tmp);
}

