/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* camera.c
   This file contains all the routines for making and modifying cameras */

#include <stdio.h>
#include <malloc.h>
#include <math.h>

#include "camera.h"
#include "world.h"
#include "vector.h"
#include "map.h"
#include "text.h"
#include "tools.h"

#include "round.h"
#include "game.h"

static Camera *allocate_camera(void);
static int free_camera(Camera *camera);

/* Allocate space for a camera */
static Camera *allocate_camera(void)
{
  Camera *camera;

  camera = malloc(sizeof(Camera));
  if(!camera)
  {
    Neprintf("Error: Unable to allocate space for camera\n");
    return NULL;
  }

  camera->world = NULL;
  camera->game = NULL;
  camera->round = NULL;

  camera->target_pos = NULL;
  camera->actual_target_pos[0] = 0.0;
  camera->actual_target_pos[1] = 0.0;
  camera->actual_target_pos[2] = 0.0;
  camera->pos[0] = 0.0;
  camera->pos[1] = 0.0;
  camera->pos[2] = 0.0;
  camera->underwater = FALSE;

  camera->rot[0] = 0.0;
  camera->rot[1] = 0.0;
  camera->rot[2] = 0.0;

  camera->alt = 0.0;
  camera->dist = 50.0;
  camera->fixed_alt = FALSE;

  /* Shudder */
  camera->shudder_percent = 0.0;
  camera->shudder = 0.0;
  camera->shudder_vector[0] = 0.0;
  camera->shudder_vector[1] = 0.0;
  camera->shudder_vector[2] = 0.0;
  camera->shudder_vector_valid = FALSE;
  camera->shudder_pos[0] = 0.0;
  camera->shudder_pos[1] = 0.0;
  camera->shudder_pos[2] = 0.0;

  return camera;
}

/* Add a camera into a round */
/* TODO: Make this support multiple cameras */
Camera *add_camera(Round *round)
{
  Camera *camera;

  if(!round || round->camera)
    return NULL;

  camera = round->camera = allocate_camera();

  camera->world = round->game->world;
  camera->game = round->game;
  camera->round = round;

  return camera;
}

/* Sets the cameras target */
void set_camera_target(Camera *camera, float *target)
{
  camera->target_pos = target;
}

/* Change the position of the camera */
void move_camera(Camera *camera, float x, float y, float z)
{
  camera->pos[0] += x;
  camera->pos[2] += z;

  if(camera->fixed_alt)
  {
    camera->alt += y;
    camera->pos[1] = get_height(camera->world->game->map,
				camera->pos[0],
				camera->pos[2]) + camera->alt;
  }
  else
    camera->pos[1] += y;
}

/* Same as above but use a pointer */
void move_camerav(Camera *camera, float *displacement)
{
  move_camera(camera,
	      displacement[0],
	      displacement[1],
	      displacement[2]);
}

/* Change the rotation on the camera */
void rotate_camera(Camera *camera, float h, float p, float r)
{
  float *rot = camera->rot;

  rot[0] += h;
  rot[1] += p;
  rot[2] += r;

  while(rot[0] > 360.0)
    rot[0] -= 360.0;
  while(rot[0] < 0.0)
    rot[0] += 360.0;  
  if(rot[1] > 90.0)
    rot[1] = 90.0;
  if(rot[1] < -90.0)
    rot[1] = -90.0;
  while(rot[2] > 360.0)
    rot[2] -= 360.0;
  while(rot[2] < 0.0)
    rot[2] += 360.0;

  update_camera_pos(camera);
}

/* Change the zoom (actually the distance from the target) */
void zoom_camera(Camera *camera, float dr)
{
  camera->dist += dr;
  if(camera->dist < 0)
    camera->dist = 0;

  update_camera_pos(camera);
}

/* Moves, shakes etc the cameras around */
void update_cameras(Round *round, float dt)
{
  Camera *camera = round->camera;

  /*while(camera)
  {*/
    update_camera(camera, dt);
  /*}*/
}

/* Update a camera */
void update_camera(Camera *camera, float dt)
{
  float shudder_disp[3];
  int i;

  if(camera->target_pos == NULL)
    return;

  for(i = 0; i < 3; i++)
    camera->actual_target_pos[i] += 0.75 * dt
      * (camera->target_pos[i] - camera->actual_target_pos[i]);

  /* Shake the camera */
  if(camera->shudder > 0.1)
  {
    /* Reduce the shudder */
    camera->shudder *= exp(-dt);

    /* Direction of shake */
    if(camera->shudder_vector_valid == FALSE)
      _make_random_vector(camera->shudder_vector);

    /* How far through that lurch we are */
    camera->shudder_percent += 5.0 * dt;
    if(camera->shudder_percent > 1.0)
      camera->shudder_percent = 1.0;

    /* The actual displacement caused by the shudder */
    _vector_scale(0.2 * camera->shudder_percent * camera->shudder,
		  camera->shudder_vector, shudder_disp);
    camera->shudder_pos[0] = shudder_disp[0];
    camera->shudder_pos[1] = shudder_disp[1];
    camera->shudder_pos[2] = shudder_disp[2];

    /* Shake in another direction next time */
    if(camera->shudder_percent == 1.0)
    {
      camera->shudder_vector_valid = FALSE;
    }
  }
  else
  {
    camera->shudder = 0.0;
    camera->shudder_pos[0] = 0.0;
    camera->shudder_pos[1] = 0.0;
    camera->shudder_pos[2] = 0.0;
  }

  update_camera_pos(camera);
}

/* Calculate the location of the camera */
void update_camera_pos(Camera *camera)
#define MIN_HEIGHT 2.0
{
  float sin_heading, cos_heading, sin_pitch, cos_pitch;
  float *pos = camera->pos, *rot = camera->rot, *tpos = camera->actual_target_pos;
  float h;

  sin_heading = sin((M_PI / 180.0) * rot[0]);
  cos_heading = cos((M_PI / 180.0) * rot[0]);
  sin_pitch = sin((M_PI / 180.0) * rot[1]);
  cos_pitch = cos((M_PI / 180.0) * rot[1]);

  /* Calculate the position (rotated around the target) */
  pos[0] = cos_pitch * sin_heading * camera->dist + tpos[0];
  pos[1] = camera->dist * -sin_pitch + tpos[1];
  pos[2] = cos_pitch * cos_heading * camera->dist + tpos[2];

  /* Don't go under the map */
  h = get_height(camera->world->game->map, pos[0], pos[2]);
  if(pos[1] < (h + MIN_HEIGHT))
    pos[1] = h + MIN_HEIGHT;

  /* Check if underwater */
  if(pos[1] < camera->world->game->effective_sea_level)
    camera->underwater = TRUE;
  else
    camera->underwater = FALSE;
}

/* Remove a camera */
int remove_camera(Camera *camera)
{
  if(!camera)
    return FALSE;

  camera->round->camera = NULL;
  free_camera(camera);

  return TRUE;
}

/* Free a camera */
static int free_camera(Camera *camera)
{
  if(!camera)
    return FALSE;

  Nfree(camera);

  return TRUE;
}
