/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* draw.c
   This file contains the routines for organising the display, the
   specific drawing functions are split into the relevant files */

#include <stdio.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <malloc.h>
#include <string.h>
#include <math.h>

#include "draw.h"
#include "world.h"
#include "text.h"
#include "draw_map.h"
#include "2d.h"
#include "input.h"
#include "about.h"
#include "player.h"
#include "player_config.h"
#include "player_list.h"
#include "projectile.h"
#include "trails.h"
#include "explosion.h"
#include "smoke.h"
#include "objects.h"
#include "start.h"
#include "cursor.h"
#include "round.h"
#include "videoc.h"
#include "shade.h"
#include "bird.h"
#include "game.h"
#include "game_config.h"
#include "round_config.h"
#include "keymap.h"
#include "glut_interface.h"
#include "display_stack.h"

#define FIELD_OF_VIEW 60.0
#define NEAR 0.1
#define FAR 10000.0

static float view_x_ratio = 1.0, view_y_ratio = 1.0;

/* Redo the window viewport */
void reshape(World *world, int w, int h)
{
  GLdouble aspect;

  aspect = (GLdouble)w / (GLdouble)h;

  glViewport(0, 0, (GLsizei) w, (GLsizei) h);

  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective(FIELD_OF_VIEW, aspect, NEAR, FAR);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  view_y_ratio = tan(0.5 * FIELD_OF_VIEW * M_PI / 180.0);
  view_x_ratio = aspect * view_y_ratio;

  /* Store this size */
  world->size_x = (int)w;
  world->size_y = (int)h;
}

/* Reset the perspective matrix */
void reset_perspective()
{
  GLint viewport[4];

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  glGetIntegerv (GL_VIEWPORT, viewport);
  gluPerspective(FIELD_OF_VIEW, (GLfloat) viewport[2]/(GLfloat) viewport[3], NEAR, FAR);

  view_y_ratio = tan(0.5 * FIELD_OF_VIEW * M_PI / 180.0);
  view_x_ratio = (GLfloat)viewport[2] / (GLfloat)viewport[3] * view_y_ratio;

  glMatrixMode(GL_MODELVIEW);
}

/* Draw the scene */
void display(World *world)
{
  NatowDisplayMode display_mode = get_display_mode(world);

  /* Start with a clean slate */
  glClearColor(0.0, 0.0, 0.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  /* Special: If shade mode down, dont change the
              display mode until the shading is over */
  if(world->shade_mode == NATOW_SHADE_DOWN)
    display_mode = NATOW_DISPLAY_UNKNOWN;

  /* Decide what to actually display */
  switch(display_mode)
  {
  case NATOW_DISPLAY_UNKNOWN:
    break;
  case NATOW_DISPLAY_MAIN_MENU:
    draw_start(world);
    draw_cursor(world);
    break;
  case NATOW_DISPLAY_KEY_CONFIG:
    draw_keymap(world, -1);
    draw_cursor(world);
    break;
  case NATOW_DISPLAY_VIDEO_CONFIG:
    draw_videoc(world, -1);
    draw_cursor(world);
    break;
  case NATOW_DISPLAY_GAME_CONFIG:
    draw_game_config(world->game);
    draw_cursor(world);
    break;
  case NATOW_DISPLAY_ABOUT:
    draw_about(world);
    draw_cursor(world);
    break;
  case NATOW_DISPLAY_PLAYER_CONFIG:
    draw_player_config(world->game);
    draw_cursor(world);
    break;
  case NATOW_DISPLAY_PLAYER_SHOP:
    draw_player_shop(world->game);
    draw_cursor(world);
    break;
  case NATOW_DISPLAY_START_ROUND:
    draw_start_round(world->game);
    draw_cursor(world);
    break;
  case NATOW_DISPLAY_ROUND:
    if(world->game)
    {
      draw_round(world->game->round);
      draw_overlay(world->game->round);
    }
    break;
  case NATOW_DISPLAY_ROUND_MENU:
    if(world->game)
    {
      draw_round(world->game->round);
      draw_round_menu(world->game->round);
    }
    draw_cursor(world);
    break;
  case NATOW_DISPLAY_ROUND_PAUSED:
    if(world->game)
      draw_round(world->game->round);

    activate_2d();
    draw_2d_text_center("Paused", 1, 98, 45, 10, TRUE);
    deactivate_2d();
    draw_cursor(world);
    break;
  case NATOW_DISPLAY_END_ROUND:
    if(world->game)
      draw_end_round(world->game->round);
    draw_cursor(world);
    break;
  case NATOW_DISPLAY_END_GAME:
    draw_end_game(world->game);
    draw_cursor(world);
    break;
  }

  /* special: if shade mode down, draw it */
  if ((world->shade_mode == NATOW_SHADE_DOWN)||(world->shade_mode == NATOW_SHADE_UP))
    draw_shade(world);

  /* Finished drawing, display now */
  swap_buffers();
}

/* Draw pincushion over surface */
void draw_pincushion(Round *round)
{
  Map *map = round->game->map;
  int dx, dy;
  float c[4] = {1.0f, 0.0f, 0.0f, 1.0f};

  glColor3f(c[0], c[1], c[2]);
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, c);

  glBegin(GL_LINES);
  for(dy = 0; dy < map->size_y; dy++)
  {
    for(dx = 0; dx < map->size_y; dx++)
    {
      glVertex3f(dx * map->x_resolution,
		 map->pos[dx][dy][1],
		 dy * map->y_resolution);
      glVertex3f(dx * map->x_resolution + map->norm[dx][dy][0] * 10,
		 map->pos[dx][dy][1] + map->norm[dx][dy][1] * 10,
		 dy * map->y_resolution + map->norm[dx][dy][2] * 10);
    }
  }
  glEnd();
}

/* Draws a grid on the plane of zero elevation */
void draw_grid(Round *round)
{
  Map *map = round->game->map;
  unsigned int x, y;
  float c[3] = {1.0f, 0.0f, 0.0f};

  glColor3f(c[0], c[1], c[2]);
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, c);
  glBegin(GL_LINES);

  for(x = 0; x < map->size_x; x++)
  {
    glVertex3f(map->x_resolution * x, 0.0f, 0.0f);
    glVertex3f(map->x_resolution * x, 0.0f, map->y_resolution * (map->size_y - 1));
  }

  for(y = 0; y < map->size_y; y++)
  {
    glVertex3f(0.0f, 0.0f, map->y_resolution * y);
    glVertex3f(map->x_resolution * (map->size_x - 1), 0.0f, map->y_resolution * y);
  }

  glEnd();
}

/* Draw the overlay */
int draw_overlay(Round *round)
#define MAX_LENGTH 50
{
  World *world;
  Game *game;
  Player *current_player = NULL;
  char text[MAX_LENGTH];
  float cash_timer;
  float ll = 0.;

  if(!round)
    return FALSE;

  world = round->world;
  game = round->game;

  activate_2d();
  set_2d_transparency(1.0);

  sprintf(text, "%.1f fps", world->fps);
  /*
  if(world->max_dt < 100.0f)
    snprintf(text, MAX_LENGTH, "%.1f fps (time lag)", world->fps);*/
  draw_2d_text(text, 1, 1, 2, FALSE);

  /* In SINGLE_PLAYER mode, draw state of all players */
  if(game->mode == SINGLE_PLAYER)
    {
    set_2d_transparency(.8);
    ll = 1.;
    current_player = world->player_list->next;
    while(current_player != NULL)
      {
      sprintf(text, "%s: %.2f", current_player->name, current_player->health);
      draw_2d_text_right(text, 1, 98, ll, -2, FALSE);
      ll += 3.3;
      current_player = current_player->next;
      }
    }

  /* Draw info on the current player (if there is one) */
  if(!(current_player = round->current_player))
  {
    deactivate_2d();
    return FALSE;
  }

  cash_timer = current_player->cash_display_timer;

  draw_2d_text(current_player->name, 1, 96, -3, FALSE);

  if((current_player->engine_on) && (game->mode == TURN_BASED))
  {
    snprintf(text, MAX_LENGTH,
	     "Acceleration = %.1f ms^-2, Speed %.1f ms^-1",
	     current_player->accel,
	     current_player->speed);
    draw_2d_text(text, 1, 92, -4, FALSE);

    snprintf(text, MAX_LENGTH,
	     "Engine engaged: %.2fs remaining",
	     current_player->drive_time);
    draw_2d_text(text, 1, 88, -4, FALSE);
  }

  if ((!current_player->engine_on && game->mode == TURN_BASED) ||
      (world->game->mode == SINGLE_PLAYER))
  {
    snprintf(text, MAX_LENGTH,
	     "Muzzle Velocity = %.1f ms^-1, Thrust %.1f ms^-2",
	     current_player->muzzle_vel,
	     current_player->thrust);
    draw_2d_text(text, 1, 92, -4, FALSE);

    snprintf(text, MAX_LENGTH, "%s (%d)",
	   current_player->selected_weapon->munition->name,
	     current_player->selected_weapon->number);
    draw_2d_text(text, 1, 88, -4, FALSE);
  }

  snprintf(text, MAX_LENGTH, "Health: %.2f%%", current_player->health);
  draw_2d_text(text, 1, 84, -4, TRUE);

  /* If a player is increasing their cash */
  if(cash_timer)
  {
#define TOTAL_DISPLAY_TIME 8.0
#define FADE_IN_TIME 2.0
#define FADE_OUT_TIME 2.0
    /* Fade in and out */
    if(cash_timer < FADE_IN_TIME)
      set_2d_transparency(cash_timer / FADE_IN_TIME);
    else if(cash_timer > FADE_OUT_TIME)
      set_2d_transparency((TOTAL_DISPLAY_TIME - cash_timer) / FADE_OUT_TIME);

    snprintf(text, MAX_LENGTH, "$%d", current_player->temp_cash);
    draw_2d_text(text, 80, 95, 0, FALSE);
    snprintf(text, MAX_LENGTH, "+$%d", current_player->temp_winnings);
    draw_2d_text(text, 80, 90, -2, FALSE);
  }

  deactivate_2d();

  glPushMatrix();
  glLoadIdentity();
  glTranslatef(0.7 * 20.0 * view_x_ratio, -0.8 * 20.0 * view_y_ratio, -20.0);
  set_player_tank_variables(current_player);
  glassDrawObject(current_player->tank->model);
  glPopMatrix();

  return TRUE;
}
