/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* file.c
   Contains common functions to access files */

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "file.h"
#include "tools.h"

/* Returns a word from the file, ie removes whitespace and ignores comments (#) */
char *get_file_word(FILE *file)
#define MAX_WORD_LENGTH 1024
{
  char word[MAX_WORD_LENGTH + 1];
  int i = 0, input_char, is_word;

  do {
    /* Clear whitespace */
    do
      input_char = fgetc(file);
    while(isspace(input_char));
    is_word = 1;

    /* Remove comments */
    if(input_char == '#')
    {
      while((input_char != EOF) && (input_char != '\n'))
	input_char = fgetc(file);
      is_word = 0;
    }
  } while(!is_word); /* Repeat until whitespace has been cleared and no comment is present */

  if(input_char == EOF)
    return NULL;

  /* Get words (eg "Eat me") */
  if(input_char == '\"')
  {
    do
    {
      input_char = fgetc(file);
      word[i] = (char) input_char;
      i++;
    } while((input_char != '\"') && (input_char != EOF) && (i < MAX_WORD_LENGTH));
    word[i-1] = '\0';
  }
  else /* Get a word surrounded by whitespace */
  {
    while(!isspace(input_char) && (input_char != EOF) && (i < MAX_WORD_LENGTH))
    {
      word[i] = (char) input_char;
      input_char = fgetc(file);
      i++;
    }
    word[i] = '\0';
  }

  return strdup(word);
}

/* Covert a file name (fname2) relative to fname1, to a filename relative to the
   current path. The returned name is allocated */
char *make_absolute_fname(const char *fname1, const char *fname2)
#define MAX_FNAME 100
{
  char new_fname[MAX_FNAME];
  char *dir = strdup(fname1);
  int i, fname1_len = strlen(fname1);

  for(i = fname1_len - 2; i >= 0; i--)
    if(dir[i] == '/') /* !!note: will have to change this for win32 */
    {
      dir[i+1] = '\0';
      break;
    }

  if(i < 0) /* Is locally relative anyway */
    snprintf(new_fname, MAX_FNAME, "%s", fname2);
  else
    snprintf(new_fname, MAX_FNAME, "%s%s", dir, fname2);

  Nfree(dir);
  return strdup(new_fname);
}

/* Return in_fname2 relative to in_fname1 */
char *make_relative_fname(const char *in_fname1, const char *in_fname2)
{
  char *full_fname1, *full_fname2;
  char *fname1, *fname2, *t;
  char final_fname[300];
  int i, last_slash, max, ndirs;

  /* Change the absolute filename to a relative one */

  /* Copy the file names */
  full_fname1 = strdup(in_fname1);
  full_fname2 = strdup(in_fname2);
  fname1 = full_fname1;
  fname2 = full_fname2;
  /* Remove the root '/'s */
  if(*fname1 == '/')
    fname1++;
  if(*fname2 == '/')
    fname2++;

  /* Remove the file from the fname1 */
  for(t = &fname1[strlen(fname1)]; t != fname1 && *t != '/'; t--);
  *(t+1) = '\0';
  
  /* Remove as many shared dirs as possible */
  last_slash = -1;
  max = strlen(fname1) > strlen(fname2) ?
    strlen(fname2) : strlen(fname1);
  for(i = 0; i < max && fname1 != '\0' && (fname1[i] == fname2[i]); i++)
  {
    if(fname1[i] == '/' && fname2[i] == '/')
      last_slash = i;
  }
  fname1 = &fname1[last_slash+1];
  fname2 = &fname2[last_slash+1];
  
  /* Add as many ..'s are required */
  ndirs = 0;
  for(i = 0; i < strlen(fname1); i++)
    if(fname1[i] == '/')
      ndirs++;

  final_fname[0] = '\0';
  for(i = 0; i < ndirs; i++)
    strncat(final_fname, "../", 300);
  
  /* Add the filename and its unique dirs to the end */
  strncat(final_fname, fname2, 300);
  
  free(full_fname1);
  free(full_fname2);
  
  return strdup(final_fname);
}
