/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* input.c
   Contains all the input routines */

#include <stdio.h>
#include <string.h>

#include "input.h"
#include "world.h"

#include "tools.h"
#include "file.h"
#include "text.h"

#include "player.h"
#include "player_config.h"
#include "player_list.h"

#include "camera.h"
#include "projectile.h"
#include "videoc.h"
#include "start.h"
#include "2d.h"
#include "ai.h"

#include "game.h"
#include "game_config.h"
#include "round.h"
#include "round_config.h"

#include "about.h"
#include "draw.h"
#include "config.h"
#include "keymap.h"
#include "glut_interface.h"
#include "display_stack.h"

/* Deal with key events */
void keyboard(World *world, int key, int x, int y)
{
  /* Decide what to actually display */
  switch(get_display_mode(world))
  {
  case NATOW_DISPLAY_UNKNOWN:
    break;
  case NATOW_DISPLAY_MAIN_MENU:
    manage_main_menu_input(world, key);
    break;
  case NATOW_DISPLAY_KEY_CONFIG:
    manage_edit_keymap_input(world, key);
    break;
  case NATOW_DISPLAY_VIDEO_CONFIG:
    manage_video_config_input(world, key);
    break;
  case NATOW_DISPLAY_GAME_CONFIG:
    manage_game_config_input(world->game, key);
    break;
  case NATOW_DISPLAY_ABOUT:
    manage_about_input(world, key);
    break;
  case NATOW_DISPLAY_PLAYER_CONFIG:
    manage_player_config_input(world->game, key);
    break;
  case NATOW_DISPLAY_PLAYER_SHOP:
    manage_player_shop_input(world->game, key);
    break;
  case NATOW_DISPLAY_START_ROUND:
    manage_start_round_input(world->game, key);
    break;
  case NATOW_DISPLAY_ROUND:
    if(world->game)
      manage_round_input(world->game->round, key);
    break;
  case NATOW_DISPLAY_ROUND_MENU:
    if(world->game)
      manage_round_menu_input(world->game->round, key);
    break;
  case NATOW_DISPLAY_ROUND_PAUSED:
    manage_round_paused_input(world, key);
    break;
  case NATOW_DISPLAY_END_ROUND:
    if(world->game)
      manage_end_round_input(world->game->round, key);
    break;
  case NATOW_DISPLAY_END_GAME:
    manage_end_game_input(world->game, key);
    break;
  }

  /* Update the display */
  post_redisplay();
}

/* Manage the key input when the round is paused */
int manage_round_paused_input(World *world, int key)
{
  /* if paused, the only available key is to unpause */
  if(key == KL(KC_pause))
    pop_display_mode(world);

  return TRUE;
}

/* Called when the mouse motion event occurs */
void motion(World *world, int x, int y, int dx, int dy)
{
  Camera *camera;
  int *button_states = get_button_states();

  if(!world->game || !world->game->round)
    return;

  camera = world->game->round->camera;

  /* not during pause */
  if (get_display_mode(world) == NATOW_DISPLAY_ROUND_PAUSED)
    return;

  if(button_states[LEFT_BUTTON] == MOUSE_DOWN)
  {
    rotate_camera(camera, -dx, 0.0, 0.0);
    rotate_camera(camera, 0.0, -dy, 0.0);
  }
  else if(button_states[RIGHT_BUTTON] == MOUSE_DOWN)
    zoom_camera(camera, (float)dy);

  post_redisplay();
}

/* Manage generic mouse events */
void mouse_event(World *world, int button, int state, int x, int y)
{
  int *button_states = get_button_states();

  /* Cursor must be inside the screen */
  world->cursor_inside = GLUT_ENTERED;

  /* if not in round, simulates a key-press action (KEY_MOUSE) */
  if (get_display_mode(world) != NATOW_DISPLAY_ROUND)
  {
    if(button_states[LEFT_BUTTON] == MOUSE_DOWN)
      keyboard(world, KEY_MOUSE1, x, y);
    else if(button_states[MIDDLE_BUTTON] == MOUSE_DOWN)
      keyboard(world, KEY_MOUSE2, x, y);
    else if(button_states[RIGHT_BUTTON] == MOUSE_DOWN)
      keyboard(world, KEY_MOUSE3, x, y);
    
    return;
  }
}

/* during game, leaving the screen -> pause the game */
void enter_leave(World *world, int state)
{
  world->cursor_inside = state;
}
