/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* keymap.c
   Manage modifying and using the keymap */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "keymap.h"

#include "tools.h"
#include "file.h"
#include "text.h"

#include "config.h"
#include "2d.h"
#include "draw.h"
#include "plasma.h"
#include "display_stack.h"
#include "glut_interface.h"

/* The initial key list */
static KeyEntry key_list[KEYS_NUMBER] =
  {
  {"start_engine",
   "Go into drive mode",
   KEY_m},
  {"accel+",
   "Accelerate tank",
   KEY_w},
  {"accel++",
   "Accelerate tank fast",
   KEY_W},
  {"accel-",
   "Decelerate tank",
   KEY_s},
  {"accel--",
   "Decelerate tank fast",
   KEY_S},
  {"brake",
   "Apply the tanks brakes",
   KEY_c},
  {"heading+",
   "Turn left",
   KEY_a},
  {"heading++",
   "Hard turn left",
   KEY_A},
  {"heading-",
   "Turn right",
   KEY_d},
  {"heading--",
   "Hard turn right",
   KEY_D},
  {"muzzle_vel+",
   "Increase muzzle velocity",
   KEY_t},
  {"muzzle_vel-",
   "Decrease muzzle velocity",
   KEY_g},
  {"missile_thrust+",
   "Increase missile thrust",
   KEY_PLUS},
  {"missile_thrust-",
   "Decrease missile thrust",
   KEY_MINUS},
  {"remove_projectile",
   "Force removal of last projectile",
   127},
  {"turret_head+",
   "Slowly rotate turret left",
   KEY_q},
  {"turret_head++",
   "Rotate turret left",
   KEY_Q},
  {"turret_head-",
   "Slowly rotate turret right",
   KEY_e},
  {"turret_head--",
   "Rotate turret right",
   KEY_E},
  {"turret_pitch-",
   "Slowly reduce turret pitch",
   KEY_r},
  {"turret_pitch--",
   "Reduce turret pitch",
   KEY_R},
  {"turret_pitch+",
   "Slowly increase turret pitch",
   KEY_f},
  {"turret_pitch++",
   "Increase turret pitch",
   KEY_F},
  {"player_finished",
   "Finish player's turn",
   KEY_SPACE},
  {"menu",
   "Access menu screen",
   KEY_ESC},
  {"keyconfig",
   "Configure game keymap",
   KEY_F1},
  {"pause",
   "Pause the game",
   KEY_p},
  {"vidconfig",
   "Configure the video settings",
   KEY_F2},
  {"next_weapon",
   "Cycle to the next weapon",
   KEY_TAB},
  {NULL,
   NULL,  /* last entry */
   0},
};

/* Get the keymap */
KeyEntry *get_keymap(void)
{
  return key_list;
}

/* Find a key by name */
KeyEntry *find_key(const char *name)
{
  int i;

  for(i = 0; key_list[i].name != NULL && strcmp(key_list[i].name, name); i++);

  if(key_list[i].name == NULL)
    return NULL;
  else
    return &key_list[i];
}

/* Convert a keycode to a key name */
char *convert_key_to_name(int key)
#define MAX_NAME_LENGTH 10
{
  static char key_name[MAX_NAME_LENGTH];

  if(key > KEY_SPACE+1 && key < 126)
  {
    key_name[0] = (char)key;
    key_name[1] = '\0';
    return key_name;
  }

  switch(key)
  {
  case KEY_SPACE:
    snprintf(key_name, MAX_NAME_LENGTH, "space");
    break;
  case KEY_TAB:
    snprintf(key_name, MAX_NAME_LENGTH, "tab");
    break;
  case KEY_BACKSPACE:
    snprintf(key_name, MAX_NAME_LENGTH, "backspace");
    break;
  case KEY_RETURN:
    snprintf(key_name, MAX_NAME_LENGTH, "enter");
    break;
  case KEY_ESC:
    snprintf(key_name, MAX_NAME_LENGTH, "esc");
    break;
  case KEY_DEL:
    snprintf(key_name, MAX_NAME_LENGTH, "del");
    break;
  case KEY_F1:
  case KEY_F2:
  case KEY_F3:
  case KEY_F4:
  case KEY_F5:
  case KEY_F6:
  case KEY_F7:
  case KEY_F8:
  case KEY_F9:
  case KEY_F10:
  case KEY_F11:
  case KEY_F12:
    key_name[0] = 'F';
    if(key < KEY_F10)
    {
      key_name[1] = (char)key - KSPE + '0';
      key_name[2] = '\0';
    }
    else
    {
      key_name[1] = (char)(0.1*(key - KSPE)) + '0';
      key_name[2] = (char)key - KEY_F10 + '0';
      key_name[3] = '\0';
    }
    break;
  case KEY_LEFT:
    snprintf(key_name, MAX_NAME_LENGTH, "left");
    break;
  case KEY_UP:
    snprintf(key_name, MAX_NAME_LENGTH, "up");
    break;
  case KEY_RIGHT:
    snprintf(key_name, MAX_NAME_LENGTH, "right");
    break;
  case KEY_DOWN:
    snprintf(key_name, MAX_NAME_LENGTH, "down");
    break;
  case KEY_PGUP:
    snprintf(key_name, MAX_NAME_LENGTH, "pageup");
    break;
  case KEY_PGDN:
    snprintf(key_name, MAX_NAME_LENGTH, "pagedown");
    break;
  case KEY_HOME:
    snprintf(key_name, MAX_NAME_LENGTH, "home");
    break;
  case KEY_END:
    snprintf(key_name, MAX_NAME_LENGTH, "end");
    break;
  case KEY_INSERT:
    snprintf(key_name, MAX_NAME_LENGTH, "insert");
    break;
  default:
    snprintf(key_name, MAX_NAME_LENGTH, "??????");
    break;
  }

  return key_name;
}

/* Convert a key name to a key code */
int convert_name_to_key(const char *name)
{
  int len = strlen(name), key;

  /* Single character */
  if(len == 1)
  {
    key = (int)name[0];
    if(key > 33 && key < 126)
      return key;
  }

  /* Is it just a key code? */
  key = atoi(name);
  if(key)
    return key;

  /* Descriptive keys */
  if(!strcmp(name, "space"))
    return KEY_SPACE;
  else if(!strcmp(name, "tab"))
    return KEY_TAB;
  else if(!strcmp(name, "del"))
    return KEY_DEL;
  else if(!strcmp(name, "backspace"))
    return KEY_BACKSPACE;
  else if(!strcmp(name, "enter") || !strcmp(name, "return"))
    return KEY_RETURN;
  else if(!strcmp(name, "esc"))
    return KEY_ESC;
  else if(!strcmp(name, "left"))
    return KEY_LEFT;
  else if(!strcmp(name, "up"))
    return KEY_UP;
  else if(!strcmp(name, "right"))
    return KEY_RIGHT;
  else if(!strcmp(name, "down"))
    return KEY_DOWN;
  else if(!strcmp(name, "pageup"))
    return KEY_PGUP;
  else if(!strcmp(name, "pagedown"))
    return KEY_PGDN;
  else if(!strcmp(name, "home"))
    return KEY_HOME;
  else if(!strcmp(name, "end"))
    return KEY_END;
  else if(!strcmp(name, "insert"))
    return KEY_INSERT;
  else if((len == 2 || len == 3) && (name[0] == 'F' || name[0] == 'f'))
  {
    return KSPE + atoi(&name[1]);
  }

  return 0;
}


/* Key-map list saving. just use to generate an initial key-map */
int save_keylist(World *world)
{
  int i;
  FILE *f;

  if((f = create_config_key()) == NULL)
  {
    Neprintf("Can't create key-file.\n");
    return(0);
  }

  fprintf(f,
	  "# This is the n.a.t.o.w keymap config file\n"
	  "# See natow.sourceforge.net for details\n"
	  "#\n"
	  "# This file was automatically generated by n.a.t.o.w\n"
	  "# If you edit it, the changes will get written over!\n"
	  "#\n"
	  "# Note that the bindings are in reverse order, that\n"
	  "# is an action (e.g. brake), will have a key (e.g. c)\n"
	  "# rather than a key having an action. So only one key\n"
	  "# is possible for each action (at the moment)\n"
	  "\n");

  for(i = 0; i < KEYS_NUMBER && key_list[i].name; i++)
    fprintf(f, "bind\t%s\t%s\n", convert_key_to_name(key_list[i].key), key_list[i].name);

  fclose(f);

  return(1);
}

/* Read a key-map */
int read_keylist(World *world)
{
  FILE *f;
  int key = 0;
  char *word = NULL, *key_number = NULL, *key_name = NULL;
  KeyEntry *entry;

  Nprintf("Reading key-map file...\n");

  if((f = open_config_key()) == NULL)
   {
    Neprintf("Can't open key-file.\n");
    return FALSE;
  }

  while(TRUE)
  {
    Nfree(word);
    word = get_file_word(f);
    /* Done */
    if(word == NULL)
    {
      fclose(f);
      return TRUE;
    }

    if(strcmp(word, "bind"))
      Neprintf("Unknown option in key-config file (%s)\n", word);
    else
    {
      key_number = get_file_word(f);
      key_name = get_file_word(f);

      if(!key_number || !key_name)
      {
	Nfree(word);
	Nfree(key_number);
	Nfree(key_name);
	Neprintf("Unexpected EOF in key-config file\n");
	fclose(f);
	return FALSE;
      }

      /* The number of the key */
      key = convert_name_to_key(key_number);

      /* Find the key that does this action */
      entry = find_key(key_name);

      if(!entry)
	Neprintf("Unknown key action %s\n", key_name);
      else
	entry->key = key;

      Nfree(key_number);
      Nfree(key_name);
    }
  }

  fclose(f);

  Nprintf("Finished reading keymap file\n");

  return TRUE;
}

/* Manage key configuration */
void manage_edit_keymap_input(World *world, int key)
{
  static int selected = 0, changing = FALSE;
  static int use_key = 0, old_key = 0, confirm_reuse = FALSE;
  int i;

  /* If confirming if going to reuse a key */
  if(confirm_reuse)
  {
    if(key == KEY_ESC)
    {
      /* Revert to the old key */
      key_list[selected].key = old_key;

      /* Finished confirming */
      confirm_reuse = FALSE;
      draw_keymap(world, selected);
      changing = FALSE;
    }

    else if(key == KEY_RETURN || key == use_key) /* return or the key */
    {
      /* Remove all other uses of this key */
      for(i = 0; key_list[i].name != NULL; i++)
      {
	if((key_list[i].key == use_key) && (i != selected))
	  key_list[i].key = 0;
      }

      /* Use this key */
      key_list[selected].key = use_key;

      /* Finished confirming */
      confirm_reuse = FALSE;
      draw_keymap(world, selected);
      changing = FALSE;
    }

    return;
  }

  /* manage keys */
  if(changing)
  {
    /* check if the key is free */
    i = 0;
    while(key_list[i].name != NULL)
    {
      if((key_list[i].key == key)&&(i != selected))
      {
	use_key = key;
	confirm_reuse = TRUE;
	draw_keymap(world, 666);
	break;
      }
      i++;
    }
    if(key_list[i].name == NULL)
    {
      key_list[selected].key = key;
      changing = FALSE;
      draw_keymap(world, selected);
    }
  }

  else
  {
    if(key == KEY_ESC) /* esc */
    {
      changing = FALSE;
      confirm_reuse = FALSE;
      use_key = 0;
      old_key = 0;
      draw_keymap(world, 1000);

      pop_display_mode(world);
      /* save new keymap */
      Nprintf("(saving new key-configuration)\n");
      save_keylist(world);
      return;
    }

    else if(key == KEY_RETURN) /* return */
    {
      /* Remove the existing selection */
      old_key = key_list[selected].key;
      key_list[selected].key = 0;
      changing = TRUE;
      draw_keymap(world, 999);
    }

    else if(key == KEY_UP) /* up */
    {
      selected--;
      if(selected < 0)
        selected = 0;
      draw_keymap(world, selected);
    }

    else if(key == KEY_DOWN) /* down */
    {
      selected++;
      if(key_list[selected].name == NULL)
        selected--;
      draw_keymap(world, selected);
    }

    else if(key == KEY_BACKSPACE || key == KEY_DEL) /* delete or backspace */
    {
      key_list[selected].key = 0;
    }
  }
}

/* Draw keymap config screen */
#define MAX_VIS 5 /* Maximum number on each side of selected item */
#define ABS(v) ((v)<0?-(v):(v))
void draw_keymap(World *world, int in_selected)
{
  static int selected = 0, confirm_reuse = FALSE, modifying = FALSE;
  int i, offset, key, key_state;

  /* Reset */
  if(in_selected == 1000)
  {
    confirm_reuse = FALSE;
    modifying = FALSE;
    return;
  }

  if(in_selected != 666 && in_selected != 999)
  {
    /* Modify the selected key */
    if(in_selected >= 0)
    {
      confirm_reuse = FALSE;
      modifying = FALSE;
      selected = in_selected;
      return;
    }
  }
  else
  {
    /* Confirming whether to reuse a key */
    if(in_selected == 666)
      confirm_reuse = TRUE;
    /* Modifying the current key */
    else if(in_selected == 999)
      modifying = TRUE;
  }

  activate_2d();

  draw_plasma(world);

  draw_2d_box(1, 1, 98, 98, 0);

  set_2d_transparency(1.0);
  draw_2d_text_center("Key configuration",
                      1, 98, 85, 5, 0);

  if(confirm_reuse)
    draw_2d_text_center("Key already used, re-enter to confirm, esc to cancel",
			1, 98, 80, -3, 0);

  draw_2d_text_center("Up/Down: select key, Ret: modify key, Esc: end",
                      1, 98, 2, -2, 0);

  if(selected < MAX_VIS)
    offset = MAX_VIS - selected;
  else
    offset = 0;

  /* Display the key list */
  for(i = selected - MAX_VIS; i <= selected + MAX_VIS; i++)
  {
    key = i + offset;

    /* End of the list */
    if(key_list[key].name == NULL)
      break;
      
    /* First display the key */
    if(key == selected)
    {
      if(modifying)
	key_state = 3; /* green */
      else
	key_state = TRUE;
    }
    else
      key_state = FALSE;
    set_2d_transparency(1.0-0.1*ABS(key-selected));
    draw_2d_text(convert_key_to_name(key_list[key].key),
		 2.0, 50 + 4.1*(selected - key + offset), -1, key_state);
      
    /* Then display the description */
    set_2d_transparency(1.0-0.1*ABS(key-selected));
    draw_2d_text(key_list[key].description,
		 20.0, 50 + 4.1*(selected - key + offset), 0, key == selected);
  }

  deactivate_2d();
}
