/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* player_stats.c
   This file contains code for each player to keep track of
   their interaction history with their opponents */

#include <stdlib.h>
#include <stdio.h>

#include "player_stats.h"
#include "world.h"

#include "text.h"
#include "tools.h"

static int player_in_array(Player **player_array, Player *player);
static void add_entry(Player *player, Player *opponent);
static PlayerStat *allocate_player_stat_entry(void);
static PlayerStat *find_stat(Player *player, Player *opponent);
static void clear_stat(PlayerStat *stat);
static void free_stat(PlayerStat *stat);

/* Build a list of statistics on opponents for all the players */
void build_all_player_statistics(World *world)
{
  Player *player;

  player = world->player_list;
  while(player)
  {
    build_player_statistics(player);
    player = player->next;
  }
}

/* Build a list of statistics on opponents for a player */
void build_player_statistics(Player *player)
{
  World *world;
  Player *used_entries[100], *opp;
  PlayerStat *stat;
  int i;

  if(!player || !player->world)
    return;

  world = player->world;
  used_entries[0] = NULL;

  /* If some entries have already been made, record these
     so they're not remade */
  if(player->player_stats)
  {
    stat = player->player_stats;
    i = 0;
    while(stat)
    {
      used_entries[i++] = stat->player;
      stat = stat->next;
    }
    used_entries[i] = NULL;
  }

  /* Add entries for all players without entries */
  opp = world->player_list;
  while(opp)
  {
    if(!player_in_array(used_entries, opp))
      add_entry(player, opp);
    opp = opp->next;
  }
}

/* Return TRUE if the player is in the array */
static int player_in_array(Player **player_array, Player *player)
{
  int i = 0;

  if(!player_array || !player)
    return FALSE;

  /* Check if it is in there */
  while(player_array[i] && player_array[i] != player)
    i++;

  /* Was found */
  if(player_array[i])
    return TRUE;

  return FALSE;
}

/* Add a player stastic entry */
static void add_entry(Player *player, Player *opponent)
{
  PlayerStat *stat;

  if(!player || !opponent)
    return;

  stat = allocate_player_stat_entry();
  if(!stat)
    return;

  stat->player = opponent;

  /* Link it in */
  stat->next = player->player_stats;
  player->player_stats = stat;
  if(stat->next)
    stat->next->last = stat;
}

/* Allocate an empty player statistic entry */
static PlayerStat *allocate_player_stat_entry(void)
{
  PlayerStat *stat;

  stat = malloc(sizeof(PlayerStat));
  if(!stat)
  {
    Neprintf("Error: Unable to allocate space for player statistic entry\n");
    return NULL;
  }

  stat->player = NULL;
  stat->kills = 0;
  stat->losses = 0;
  stat->damage_taken = 0.0;
  stat->damage_inflicted = 0.0;

  stat->range = 0.0;
  stat->visible = FALSE;
  stat->closest = FALSE;
  stat->strongest = FALSE;
  stat->weakest = FALSE;

  stat->threat = 0.0;
  stat->easiness = 0.0;

  stat->last = NULL;
  stat->next = NULL;

  return stat;
}

/* Record damage to victim in players statistics */
void record_damage(Player *player, Player *victim, float damage)
{
  PlayerStat *stat;

  if(!player || !victim)
    return;

  /* Record the damage by the player */
  stat = find_stat(player, victim);
  if(stat)
    stat->damage_inflicted += damage;
  else
    Neprintf("Trying to record damage on \"%s\" by \"%s\", but \"%s\" has "
	     "no statstics on them\n", victim->name, player->name, player->name);

  /* Record the damage on the victim */
  stat = find_stat(victim, player);
  if(stat)
    stat->damage_taken += damage;
  else
    Neprintf("Trying to record damage taken by \"%s\" to \"%s\", but \"%s\" has "
	     "no statstics on them\n", player->name, victim->name, victim->name);
}

/* Record a kill in players statistics */
void record_kill(Player *player, Player *victim)
{
  PlayerStat *stat;

  if(!player || !victim)
    return;

  /* Record the kill by the killer */
  stat = find_stat(player, victim);
  if(stat)
    stat->kills++;
  else
    Neprintf("Trying to record a kill on \"%s\" by \"%s\", but \"%s\" has "
	     "no statstics on them\n", victim->name, player->name, player->name);

  /* Record the loss on the victim */
  stat = find_stat(victim, player);
  if(stat)
    stat->losses++;
  else
    Neprintf("Trying to record a loss on \"%s\" from \"%s\", but \"%s\" has "
	     "no statstics on them\n", player->name, victim->name, victim->name);
}

/* Find a stastic entry that 'player' has on 'opponent' */
static PlayerStat *find_stat(Player *player, Player *opponent)
{
  PlayerStat *stat;

  if(!player || !opponent)
    return NULL;

  stat = player->player_stats;
  while(stat)
  {
    if(stat->player == opponent)
      return stat;
    stat = stat->next;
  }

  return NULL;
}

/* Clear out the statistics the player stores on opponents */
void clear_player_stats(Player *player)
{
  PlayerStat *stat;

  if(!player)
    return;

  stat = player->player_stats;
  while(stat)
  {
    clear_stat(stat);
    stat = stat->next;
  }
}

/* Clear the statistics stored in a player statistic entry */
static void clear_stat(PlayerStat *stat)
{
  if(!stat)
    return;

  stat->kills = 0;
  stat->losses = 0;
  stat->damage_taken = 0.0;
  stat->damage_inflicted = 0.0;
}

/* Remove all statistics that a player stores on opponents */
void free_player_stats(Player *player)
{
  PlayerStat *stat, *tmp;

  if(!player)
    return;

  stat = player->player_stats;
  while(stat)
  {
    tmp = stat->next;
    free_stat(stat);
    stat = tmp;
  }

  player->player_stats = NULL;
}

/* Free a player statistic entry */
static void free_stat(PlayerStat *stat)
{
  Nfree(stat);
}
