/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* start.c
   Code for starting screen (the main menu) */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "start.h"
#include "world.h"
#include "2d.h"
#include "image.h"
#include "update.h"
#include "input.h"
#include "cursor.h"
#include "plasma.h"
#include "text.h"
#include "round.h"
#include "game.h"
#include "draw.h"
#include "tools.h"
#include "plasma.h"
#include "display_stack.h"
#include "glut_interface.h"

/* image for natow logo */
unsigned char *natow_logo=NULL;
/* corresponding texture */
GLuint t_logo=0;
int logo_x, logo_y;



void treat_glerror()
{
  Neprintf("GL status: '%s'\n", (char*)glGetString(glGetError()));
}

/* init the start-menu screen */
void init_start(World *world)
{
  Nprintf("Initializing main-menu...\n");

  regen_logo(world);

  fflush(stdout);

  Nprintf("Main-menu ready.\n");
}

/* free the start-menu screen */
void free_start(World *world)
{
  if (t_logo != 0)
    glDeleteTextures(1, &t_logo);
}

/* (Re)generate the logo texture */
void regen_logo(World *world)
{
  int tx, ty;

  if (t_logo != 0)
    glDeleteTextures(1, &t_logo);

  /* load the natow logo */
  read_2d_image_rgb("images/logos/logo.rgb", &natow_logo, &tx, &ty);

  /* create the corresponding texture */
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glGenTextures(1, &t_logo);
  glBindTexture(GL_TEXTURE_2D, t_logo);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, world->texture_mode);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, world->texture_mode);
  glTexImage2D(GL_TEXTURE_2D,
	       0,
	       4,
	       tx,
	       ty,
	       0,
	       GL_RGBA,
	       GL_UNSIGNED_BYTE,
	       (const GLvoid*)natow_logo);
  Nfree(natow_logo);
  natow_logo = NULL;
  logo_x = tx;
  logo_y = ty;
}

/* Update the properties of the logo texture */
void update_logo(World *world)
{
  glBindTexture(GL_TEXTURE_2D, t_logo);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, world->texture_mode);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, world->texture_mode);
}

/* manage key actions and selections for the main menu */
#define NOT_SELECTED -9999
void manage_main_menu_input(World *world, int key)
{
  static int mouse_selected = -1;

  /* special: if key <=0, it is an information send by the
       drawing function to inform the manager that the mouse
       is over |key| line (as it is to the drawing function
       to deal with graphical events */
  if (key <= 0)
    {
    if (key == NOT_SELECTED)
      mouse_selected = -1;
    else
      mouse_selected = nABS(key);
    return;
    }

  if(key == KEY_ESC)
  {
    free_world(world);
    Nprintf("Quiting n.a.t.o.w\n");
    exit(0); /* exit n.a.t.o.w */
  }
  else if(key == KEY_F1) /* F1 */
  {
    push_display_mode(world, NATOW_DISPLAY_KEY_CONFIG);
    return;
  }
  else if(key == KEY_F2) /* F2 */
  {
    push_display_mode(world, NATOW_DISPLAY_VIDEO_CONFIG);
    return;
  }
  else if(key == KEY_F5) /* F2 */
  {
    push_display_mode(world, NATOW_DISPLAY_ABOUT);
    return;
  }
  else if(key == KEY_SPACE || key == KEY_RETURN)
  {
    create_game(world);
    set_display_mode(world, NATOW_DISPLAY_GAME_CONFIG);
  }

  /* Mouse selection (left click) */
  if (key == KEY_MOUSE1)
  {
  switch(mouse_selected)
    {
    case 0: /* better to call ourself, to be treated properly... */
      manage_main_menu_input(world, KEY_F1);
      break;
    case 1:  /* ...without having to duplicate code */
      manage_main_menu_input(world, KEY_F2);
      break;
    case 2:
      manage_main_menu_input(world, KEY_ESC);
      break;
    case 3:
      manage_main_menu_input(world, KEY_RETURN);
      break;
    case 4:
      manage_main_menu_input(world, KEY_F5);
      break;
    }
  return;
  }
}

/* manage start-menu drawing */
void draw_start(World *world)
{
  static float scale=0.;
  static float last_time = 0.;
  float dt;
  char *text = NULL;
  static int mouse_selected = -1;
  int mouse_used = FALSE;

  activate_2d();

  /* draw the background */
  glDisable(GL_TEXTURE_2D);

  draw_plasma(world);

  dt = world->total_elapsed_time - last_time;
  last_time = world->total_elapsed_time;

  set_2d_transparency(1.);

  /* draw surrounding box */
  draw_2d_box(1, 1, 98, 98, 0);

  /* update scaling */
  
  scale = (((long)((world->total_elapsed_time*2.)*100.))%2000 - 500)/200.;
  if (scale > 2.5)
    scale = 5. - (scale);
  /* draw the NATOW logo, and let it move */
  glBindTexture(GL_TEXTURE_2D, t_logo);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glColor4f(1,1,1,1);
  glEnable(GL_TEXTURE_2D);
  glBegin(GL_QUADS);
  glTexCoord2f(0, 1);
  glVertex2f(5-scale, 95+scale);
  glTexCoord2f(0, 0);
  glVertex2f(5-scale, 95-(90.+2.*scale)*((float)logo_y/logo_x));
  glTexCoord2f(1, 0);
  glVertex2f(95+scale, 95-(90.+2.*scale)*((float)logo_y/logo_x));
  glTexCoord2f(1, 1);
  glVertex2f(95+scale, 95+scale);
  glEnd();
  glDisable(GL_TEXTURE_2D);

  /* draw general info */
  draw_2d_text_center("http://natow.sourceforge.net", 5, 90, 2, 0, 0);

  /* Draw game info */

  text = "Start Game";
  draw_2d_text_center(text, 0, 100, 43, 0, (mouse_selected == 3));
  if(cursor_inside_text())
  {
    mouse_selected = 3;
    mouse_used = TRUE;
  }

  text = "Configure Keymap";
  draw_2d_text_center(text, 0, 100, 36, 0, (mouse_selected == 0));
  if(cursor_inside_text())
  {
    mouse_selected = 0;
    mouse_used = TRUE;
  }

  text = "Configure Video";
  draw_2d_text_center(text, 0, 100, 29, 0, (mouse_selected == 1));
  if(cursor_inside_text())
  {
    mouse_selected = 1;
    mouse_used = TRUE;
  }

  text = "About";
  draw_2d_text_center(text, 0, 100, 22, 0, (mouse_selected == 4));
  if(cursor_inside_text())
  {
    mouse_selected = 4;
    mouse_used = TRUE;
  }

  text = "Quit";
  draw_2d_text_center(text, 0, 100, 15, 0, (mouse_selected == 2));
  if(cursor_inside_text())
  {
    mouse_selected = 2;
    mouse_used = TRUE;
  }

  /* Didn't select anything this turn */
  if(!mouse_used)
    mouse_selected = -1;

  /* if mouse selection, send the value to the manager */
  if (mouse_selected >= 0) /* send a false keybord entry */
    keyboard(world, -mouse_selected, 0, 0);
  else
    keyboard(world, NOT_SELECTED, 0, 0);

  deactivate_2d();
}
#undef NOT_SELECTED
