/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* texture.c
   Contains code to make openGL textures from image files */

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <GL/gl.h>

#include "texture.h"
#include "image.h"
#include "tools.h"
#include "text.h"

/* Loads a .rgb image and makes an openGL texture out of it */
GLuint load_gl_texture(const char *fname, int is_one_colour)
{
  IMAGE *texture_image = ImageLoad((char *)fname);
  GLint texture_name;
  GLubyte *texture_data;
  GLenum data_type;
  int x, y, pad = 0;

  if(!texture_image)
  {
    Neprintf("Error loading texture :%s\n", fname);
    return 0;
  }

  if(is_one_colour)
  {
    texture_data = malloc(sizeof(GLubyte) * texture_image->sizeX * texture_image->sizeY);

    for(x = 0; x < texture_image->sizeX; x++)
    {
      for(y = 0; y < texture_image->sizeY; y++)
      {
	/* Get the red component */
	texture_data[(x + y * texture_image->sizeX)]
	  = texture_image->data[(x + texture_image->sizeX * y)
				* texture_image->sizeZ + pad];
      }
      pad += (texture_image->sizeX * texture_image->sizeZ) % 4;
    }
  }
  else
  {
    /* Allocate space for the texture data */
    /* !!note should really be able to be accessed later */
    texture_data = malloc(sizeof(GLubyte)
			  * texture_image->sizeX
			  * texture_image->sizeY
			  * texture_image->sizeZ);

    /* Copy the data (from the image) to a new array */
    for(x = 0; x < texture_image->sizeX; x++)
    {
      for(y = 0; y < texture_image->sizeY; y++)
      {
	texture_data[(x + y * texture_image->sizeX) * texture_image->sizeZ]
	  = texture_image->data[(x + texture_image->sizeX * y)
				* texture_image->sizeZ + pad];
	texture_data[(x + y * texture_image->sizeX) * texture_image->sizeZ + 1]
	  = texture_image->data[(x + texture_image->sizeX * y) *
				texture_image->sizeZ + 1 + pad];
	texture_data[(x + y * texture_image->sizeX) * texture_image->sizeZ + 2]
	  = texture_image->data[(x + texture_image->sizeX * y) *
				texture_image->sizeZ + 2 + pad];
	if(texture_image->sizeZ > 3)
	  texture_data[(x + y * texture_image->sizeX) * texture_image->sizeZ + 3]
	    = texture_image->data[(x + texture_image->sizeX * y) *
				  texture_image->sizeZ + 3 + pad];
      }
      pad += (texture_image->sizeX * texture_image->sizeZ) % 4;
    }
  }

  /* Generate the texture */
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glGenTextures(1, &texture_name);
  glBindTexture(GL_TEXTURE_2D, texture_name);

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

  if(is_one_colour)
  {
    data_type = GL_LUMINANCE;
    texture_image->sizeZ = 1;
  }
  else if(texture_image->sizeZ > 3)
    data_type = GL_RGBA;
  else
    data_type = GL_RGB;
  glTexImage2D(GL_TEXTURE_2D,
	       0,
	       texture_image->sizeZ,
	       texture_image->sizeX,
	       texture_image->sizeY,
	       0,
	       data_type,
	       GL_UNSIGNED_BYTE,
	       texture_data);

  /* The data is copied so remove the local version */
  Nfree(texture_data);

  /* Free the image */
  Nfree(texture_image->data);
  Nfree(texture_image);

  return texture_name;
}

/* Make an empty natow texture */
NatowTexture *new_natow_texture(const char *fname, unsigned int width, unsigned int height)
{
  NatowTexture *natow_texture;
  int w, h;
  GLubyte ***data;

  natow_texture = malloc(sizeof(NatowTexture));
  if(natow_texture == NULL)
  {
    Neprintf("Error: unable to allocate space for natow-texture\n");
    return NULL;
  }

  /* Store the filename */
  if(fname == NULL)
    natow_texture->fname = strdup("");
  else
    natow_texture->fname = strdup(fname);

  natow_texture->width = width;
  natow_texture->height = height;

  /* Allocate enough space for the texture */
  natow_texture->data = malloc(sizeof(GLubyte **) * width);
  data = natow_texture->data;
  for(w = 0; w < width; w++)
  {
    data[w] = malloc(sizeof(GLubyte *) * height);
    for(h = 0; h < height; h++)
      data[w][h] = malloc(sizeof(GLubyte) * 3);
  }

  return natow_texture;
}

/* Free a natow texture */
void free_natow_texture(NatowTexture *tex)
{
  int w, h;

  if(tex == NULL)
    return;

  Nfree(tex->fname);

  for(w = 0; w < tex->width; w++)
  {
    for(h = 0; h < tex->height; h++)
      Nfree(tex->data[w][h]);
    Nfree(tex->data[w]);
  }
  Nfree(tex->data);
  Nfree(tex);
}

/* Load a natow-texture, and make an array for it */
NatowTexture *load_natow_texture(const char *fname)
{
  IMAGE *texture_image = ImageLoad((char *)fname);
  NatowTexture *natow_texture;
  int w, h, c, pad = 0;

  if(texture_image == NULL)
  {
    Neprintf("Error loading natow texture :\"%s\"\n", fname);
    return NULL;
  }

  natow_texture = new_natow_texture(fname, texture_image->sizeX, texture_image->sizeY);
  if(natow_texture == NULL)
  {
    /* Free the image */
    Nfree(texture_image->data);
    Nfree(texture_image);
    return NULL;
  }

  /* Copy the data over */
  for(w = 0; w < natow_texture->width; w++)
  {
    for(h = 0; h < natow_texture->height; h++)
      for(c = 0; c < 3; c++)
	natow_texture->data[w][h][c] = 
	  texture_image->data[(w + texture_image->sizeX * h)
			      * texture_image->sizeZ + pad + c];
    pad += (texture_image->sizeX * texture_image->sizeZ) % 4;
  }

  /* Free the image */
  Nfree(texture_image->data);
  Nfree(texture_image);

  return natow_texture;
}
