/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* videoc.c
   Management for video configuration screen */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "videoc.h"
#include "world.h"
#include "game.h"
#include "2d.h"
#include "file.h"
#include "text.h"
#include "draw.h"
#include "config.h"
#include "tools.h"
#include "plasma.h"
#include "map_texture.h"
#include "display_stack.h"
#include "glut_interface.h"
#include "start.h"

static char *video_texts[] =
  {"Screen resolution", "Enable Lighting", "Render Water", "Render Textures",
   "Texture Mode", "Render Sky", "Render Fog", "Use smooth shading", "Use Birds",
   "Level Of Detail ranges", "Smoke entities %", "Map texture resolution",
   "Enable Fading", "Plasma detail", NULL};

/* Set everything to low detail */
void set_low_detail(World *world)
{
  Nprintf("Setting low detail...\n");

  world->LOD_boundary1 = 15;
  world->LOD_boundary2 = 30;
  world->LOD_boundary3 = 60;
  /* The accuracy of the texture */
  /* Accurate to 5m */
  world->tex_x_resolution = 5.0;
  world->tex_y_resolution = 5.0;

  world->is_smooth = FALSE;
  world->is_fog = FALSE;
  world->is_light = FALSE;
  world->is_water = FALSE;
  world->is_texture = FALSE;
  world->is_sky = FALSE;
  world->is_bird = FALSE;
  world->is_shading = FALSE;
  world->smoke_percent = 0;
  world->plasma_detail = PLASMA_MIN;
  world->texture_mode = GL_NEAREST;

  world->video_size_x = 320;
  world->video_size_y = 240;
  world->fullscreen = FALSE;
}

/* Set everything to hi detail */
void set_hi_detail(World *world)
{
  Nprintf("Setting hi detail...\n");

  world->LOD_boundary1 = 30;
  world->LOD_boundary2 = 60;
  world->LOD_boundary3 = 90;
  /* The accuracy of the texture */
  /* Accurate to 5m */
  world->tex_x_resolution = 5.0;
  world->tex_y_resolution = 5.0;

  world->is_smooth = TRUE;
  world->is_fog = TRUE;
  world->is_light = TRUE;
  world->is_water = TRUE;
  world->is_texture = TRUE;
  world->is_sky = TRUE;
  world->is_bird = TRUE;
  world->is_shading = TRUE;
  world->smoke_percent = 200;
  world->plasma_detail = PLASMA_MAX;
  world->texture_mode = GL_LINEAR;

  /* is their a way to get the max resolution available ? */
  world->video_size_x = 640;
  world->video_size_y = 480;
  world->fullscreen = TRUE;
}

/* read the video-config file */
int read_videoc(World *world)
{
  FILE *f;
  char *option = NULL, *value = NULL, *screensize = NULL, *lod = NULL;

  Nprintf("Reading video-config file...\n");

  if((f = open_config_video()) == NULL)
  {
    Neprintf("  Unable to open video-config file.\n");
    return FALSE;
  }

  do
  {
    Nfree(option);
    option = get_file_word(f);
    Nfree(value);
    if(option != NULL)
      value = get_file_word(f);

    if(option == NULL);
    else if(!strcmp(option, "screensize"))
    {
      Nfree(screensize);
      screensize = strdup(value);
    }

    else if(!strcmp(option, "fullscreen"))
    {
      world->fullscreen = atoi(value);
    }

    else if(!strcmp(option, "lod") || !strcmp(option, "LOD"))
    {
      Nfree(lod);
      lod = strdup(value);
    }

    else if(!strcmp(option, "fog"))
    {
      world->is_fog = atoi(value);
    }

    else if(!strcmp(option, "light"))
    {
      world->is_light = atoi(value);
    }

    else if(!strcmp(option, "texture"))
    {
      world->is_texture = atoi(value);
    }

    else if(!strcmp(option, "bird"))
    {
      world->is_bird = atoi(value);
    }

    else if(!strcmp(option, "sky"))
    {
      world->is_sky = atoi(value);
    }

    else if(!strcmp(option, "smoke"))
    {
      world->smoke_percent = atoi(value);
    }

    else if(!strcmp(option, "tmode"))
    {
      world->texture_mode = (GLenum)atoi(value);
    }

    else if(!strcmp(option, "smooth"))
    {
      world->is_smooth = atoi(value);
    }

    else if(!strcmp(option, "water"))
    {
      world->is_water = atoi(value);
    }

    else if(!strcmp(option, "plasma"))
    {
      world->plasma_detail = atoi(value);
    }

    else if(!strcmp(option, "shading"))
    {
      world->is_shading = atoi(value);
    }

    else if(!strcmp(option, "tex_res"))
    {
      world->tex_x_resolution = world->tex_y_resolution = atof(value);
    }
    else
      Neprintf("Warning: Unknown option \"%s\" in video-config file\n", option);
  } while(option != NULL);

  fclose(f);

  /* Read video resolution */
  if(screensize != NULL)
    sscanf(screensize, "%d%d", &world->video_size_x, &world->video_size_y);

  /* Read LODs */
  if(lod != NULL)
    sscanf(lod, "%d%d%d",
	   &world->LOD_boundary1, &world->LOD_boundary2, &world->LOD_boundary3);

  Nprintf("Video config read\n");

  return TRUE;
}

/* Write the video-config file */
int write_videoc(World *world)
{
  FILE *f;

  if((f = create_config_video()) == NULL)
  {
    Neprintf("  Cannot write video-config file!\n");
    return FALSE;
  }

  fprintf(f,
	  "# The n.a.t.o.w video config file\n"
	  "# See natow.sourceforge.net for details\n"
	  "\n");
  fprintf(f, "screensize\t\"%d %d\"\n", world->video_size_x, world->video_size_y);
  fprintf(f, "fullscreen\t%d\n", world->fullscreen);
  fprintf(f, "fog\t\t%d\n", world->is_fog);
  fprintf(f, "light\t\t%d\n", world->is_light);
  fprintf(f, "texture\t\t%d\n", world->is_texture);
  fprintf(f, "sky\t\t%d\n", world->is_sky);
  fprintf(f, "smooth\t\t%d\n", world->is_smooth);
  fprintf(f, "water\t\t%d\n", world->is_water);
  fprintf(f, "lod\t\t\"%d %d %d\"\n",
	  world->LOD_boundary1, world->LOD_boundary2, world->LOD_boundary3);
  fprintf(f, "tmode\t\t%d\n", (int)world->texture_mode);
  fprintf(f, "smoke\t\t%d\n", world->smoke_percent);
  fprintf(f, "bird\t\t%d\n", world->is_bird);
  fprintf(f, "tex_res\t\t%f\n", world->tex_x_resolution);
  fprintf(f, "shading\t\t%d\n", world->is_shading);
  fprintf(f, "plasma\t\t%d\n", world->plasma_detail);

  fclose(f);

  return TRUE;
}

/* Increase the texture render quality */
void increase_proj(World *world)
{
  if(world->texture_mode == GL_NEAREST)
  {
    world->texture_mode = GL_LINEAR;
    return;
  }
  /*
  if(world->texture_mode == GL_LINEAR)
  {
    world->texture_mode = GL_BILINEAR;
    return;
  }
  */
}

/* Decrease the texture render quality */
void decrease_proj(World *world)
{
  if(world->texture_mode == GL_LINEAR)
  {
    world->texture_mode = GL_NEAREST;
    return;
  }
}

/* Decrease the screen resolution */
void increase_resolution(World *world)
{
  world->fullscreen = FALSE;

  switch(world->video_size_x)
  {
  case 320:
    world->video_size_x = 640;
    world->video_size_y = 480;
    break;
  case 640:
    world->video_size_x = 960;
    world->video_size_y = 720;
    break;
  case 960:
    world->fullscreen = TRUE;
    break;
  default:
    world->video_size_x = 640;
    world->video_size_y = 480;
    break;
  }
}

/* Decrease the screen resolution */
void decrease_resolution(World *world)
{
  if(world->fullscreen)
  {
    world->fullscreen = FALSE;
    world->video_size_x = 960;
    world->video_size_y = 720;
    return;
  }

  world->fullscreen = FALSE;

  switch(world->video_size_x)
  {
  case 960:
    world->video_size_x = 640;
    world->video_size_y = 480;
    break;
  case 640:
    world->video_size_x = 320;
    world->video_size_y = 240;
    break;
  case 320:
    break;
  default:
    world->video_size_x = 640;
    world->video_size_y = 480;
    break;
  }
}

/* Increase the level of detail on the map */
void increase_LOD(World *world)
{
  switch(world->LOD_boundary1)
  {
  case 10:
    world->LOD_boundary1 = 15;
    world->LOD_boundary2 = 30;
    world->LOD_boundary3 = 60;
    break;
  case 15:
    world->LOD_boundary1 = 20;
    world->LOD_boundary2 = 40;
    world->LOD_boundary3 = 80;
    break;
  case 20:
    world->LOD_boundary1 = 40;
    world->LOD_boundary2 = 80;
    world->LOD_boundary3 = 160;
    break;
  case 40:
    world->LOD_boundary1 = 100;
    world->LOD_boundary2 = 200;
    world->LOD_boundary3 = 400;
    break;
  case 100:
    break;
  default:
    world->LOD_boundary1 = 10;
    world->LOD_boundary2 = 20;
    world->LOD_boundary3 = 40;
    break;
  }
}

/* Increase the level of detail on the map */
void decrease_LOD(World *world)
{
  switch(world->LOD_boundary1)
  {
  case 15:
    world->LOD_boundary1 = 10;
    world->LOD_boundary2 = 20;
    world->LOD_boundary3 = 40;
    break;
  case 20:
    world->LOD_boundary1 = 15;
    world->LOD_boundary2 = 30;
    world->LOD_boundary3 = 60;
    break;
  case 40:
    world->LOD_boundary1 = 20;
    world->LOD_boundary2 = 40;
    world->LOD_boundary3 = 80;
    break;
  case 100:
    world->LOD_boundary1 = 40;
    world->LOD_boundary2 = 80;
    world->LOD_boundary3 = 160;
    break;
  case 10:
    break;
  default:
    world->LOD_boundary1 = 10;
    world->LOD_boundary2 = 20;
    world->LOD_boundary3 = 40;
    break;
  }
}

/* Manage the video-config menu */
void manage_video_config_input(World *world, int key)
{
  static int selected = 0, old_width = -1;
  static int tex_res_changed = FALSE;
  static GLenum old_tex_mode = GL_NONE;

  if(old_tex_mode == GL_NONE)
    old_tex_mode = world->texture_mode;

  if(old_width == -1)
    old_width = world->size_x;

  /* special: gfx send the mouse-selected entry */
  if (key < 0)
  {
    selected = -key - 1;
    return;
  }

  if(key == KEY_DOWN) /* down */
  {
    selected++;
    if(video_texts[selected] == NULL)
      selected--;
    draw_videoc(world, selected);
  }
  else if(key == KEY_UP) /* up */
  {
    selected--;
    if(selected < 0)
      selected = 0;
    draw_videoc(world, selected);
  }
  else if(key == KEY_PLUS || key == KEY_RIGHT || key == KEY_MOUSE1) /* + or right */
  {
    switch(selected)
    {
    case 0: /* resolution */
      increase_resolution(world);
      break;
    case 1: /* lighting */
      world->is_light = TRUE;
      break;
    case 2: /* water */
      world->is_water = TRUE;
      break;
    case 3: /* textures */
      world->is_texture = TRUE;
      break;
    case 4: /* texture mode */
      increase_proj(world);
      break;
    case 5: /* sky */
      world->is_sky = TRUE;
      break;
    case 6: /* fog */
      world->is_fog = TRUE;
      break;
    case 7: /* smooth shading */
      world->is_smooth = TRUE;
      break;
    case 8: /* birds */
      world->is_bird = TRUE;
      break;
    case 9: /* LOD */
      increase_LOD(world);
      break;
    case 10: /* smoke */
      world->smoke_percent += 10.0;
      if(world->smoke_percent > 990)
        world->smoke_percent = 990;
      break;
    case 11: /* tex. res. */
      world->tex_x_resolution += 0.5;
      if(world->tex_x_resolution > 10.0)
	world->tex_x_resolution = 10.0;
      world->tex_y_resolution = world->tex_x_resolution;
      tex_res_changed = TRUE;
      break;
    case 12: /* shading */
      world->is_shading = TRUE;
      break;
    case 13: /* plasma detail */
      if (world->plasma_detail < PLASMA_MAX)
        world->plasma_detail++;
      break;
    }
  }
  else if (key == KEY_MINUS || key == KEY_LEFT || key == KEY_MOUSE3) /* - or left */
  {
    switch(selected)
    {
    case 0: /* resolution */
      decrease_resolution(world);
      break;
    case 1: /* lighting */
      world->is_light = FALSE;
      break;
    case 2: /* water */
      world->is_water = FALSE;
      break;
    case 3: /* textures */
      world->is_texture = FALSE;
      break;
    case 4: /* texture mode */
      decrease_proj(world);
      break;
    case 5: /* sky */
      world->is_sky = FALSE;
      break;
    case 6: /* fog */
      world->is_fog = FALSE;
      break;
    case 7: /* smooth shading */
      world->is_smooth = FALSE;
      break;
    case 8: /* birds */
      world->is_bird = FALSE;
      break;
    case 9: /* LOD */
      decrease_LOD(world);
      break;
    case 10: /* smoke */
      world->smoke_percent -= 10.0;
      if(world->smoke_percent < 0)
        world->smoke_percent = 0;
      break;
    case 11: /* tex. res. */
      world->tex_x_resolution -= 0.5;
      if(world->tex_x_resolution < 0.5)
	world->tex_x_resolution = 0.5;
      world->tex_y_resolution = world->tex_x_resolution;
      tex_res_changed = TRUE;
      break;
    case 12: /* shading */
      world->is_shading = FALSE;
      break;
    case 13: /* plasma detail */
      if (world->plasma_detail > PLASMA_MIN)
        world->plasma_detail--;
      break;
    }
  }
  else if (key == KEY_ESC)
  {
    if(tex_res_changed && world->game && world->game->map)
    {
      Nprintf("Recalculating map surface texture\n");
      generate_map_texture(world->game->map);
    }

    /* Change texture mode on textures */
    if(old_tex_mode != world->texture_mode)
    {
      regen_logo(world);
      update_boxs(world);
      update_fonts(world);
      regen_plasma(world);

      if(world->game && world->game->map)
      {
	glBindTexture(GL_TEXTURE_2D, world->game->map->texture);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, world->texture_mode);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, world->texture_mode);
      }
    }

    /* Reshape the window */
    if(old_width != world->video_size_x)
    {
      Nprintf("Resizing window\n");
      if(world->fullscreen)
	fullscreen_glut();
      else
      {
	reshape_glut_window(world->video_size_x, world->video_size_y);
	move_glut_window(0, 0);
      }
    }

    if(world->is_light)
      glEnable(GL_LIGHTING);
    else
      glDisable(GL_LIGHTING);

    if(world->is_smooth)
      glShadeModel(GL_SMOOTH);
    else
      glShadeModel(GL_FLAT);

    /* Set variables back to defaults */
    tex_res_changed = FALSE;
    old_tex_mode = GL_NONE;
    old_width = -1;

    /* Go back to the old display */
    pop_display_mode(world);

    /* Save new video configuration */
    Nprintf("(saving new video-configuration)\n");
    write_videoc(world);
  }
}

/* Draw the video-config screen */
void draw_videoc(World *world, int in_selected)
#define MAX_VIS 5 /* Maximum number on each side of selected item */
#define MAX_VALUE_LENGTH 12
{
  static int selected = 0;
  char value_text[MAX_VALUE_LENGTH];
  int i, offset, index, tf_value, to_change = -1;

  /* Reset */
  if(in_selected == 99)
  {
    selected = 0;
    return;
  }

  /* Modify the selected key */
  if(in_selected >= 0)
  {
    selected = in_selected;
    return;
  }

  activate_2d();

  draw_plasma(world);

  draw_2d_box(1, 1, 98, 98, 0);

  set_2d_transparency(1.0);
  draw_2d_text_center("Video configuration",
                      1, 98, 85, 5, 0);

  draw_2d_text_center("Up/Down: select, +/-: change value, Esc: end",
                      1, 98, 2, -2, 0);

  /* If in the first few, don't have the selected
     one in the middle of the screen */
  if(selected < MAX_VIS)
    offset = MAX_VIS - selected;
  else
    offset = 0;

  to_change = -1;
  /* Display the key list */
  for(i = selected - MAX_VIS; i <= selected + MAX_VIS; i++)
  {
    tf_value = -1; /* A true/false value */
    index = i + offset;

    if(index < 0 || video_texts[index] == NULL)
      break;

    switch(index)
    {
    case 0: /* resolution */
      if(world->fullscreen)
	snprintf(value_text, MAX_VALUE_LENGTH, "Fullscreen");
      else
	snprintf(value_text, MAX_VALUE_LENGTH, "%dx%d",
		 world->video_size_x, world->video_size_y);
      break;
    case 1: /* lighting */
      tf_value = world->is_light;
      break;
    case 2: /* water */
      tf_value = world->is_water;
      break;
    case 3: /* textures */
      tf_value = world->is_texture;
      break;
    case 4: /* texture mode */
      switch(world->texture_mode)
      {
      case GL_NEAREST:
	snprintf(value_text, MAX_VALUE_LENGTH, "Nearest");
	break;
      case GL_LINEAR:
	snprintf(value_text, MAX_VALUE_LENGTH, "Linear");
	break;
      default:
	snprintf(value_text, MAX_VALUE_LENGTH, "??????");
	break;
      }
      break;
    case 5: /* sky */
      tf_value = world->is_sky;
      break;
    case 6: /* fog */
      tf_value = world->is_fog;
      break;
    case 7: /* smooth shading */
      tf_value = world->is_smooth;
      break;
    case 8: /* birds */
      tf_value = world->is_bird;
      break;
    case 9: /* LOD */
      snprintf(value_text, MAX_VALUE_LENGTH, "%d/%d/%d",
	       world->LOD_boundary1, world->LOD_boundary2, world->LOD_boundary3);
      break;
    case 10: /* smoke */
      snprintf(value_text, MAX_VALUE_LENGTH, "%d%%", world->smoke_percent);
      break;
    case 11: /* tex. res. */
      snprintf(value_text, MAX_VALUE_LENGTH, "%.2fm", world->tex_x_resolution);
      break;
    case 12: /* shading */
      tf_value = world->is_shading;
      break;
    case 13: /* plasma detail */
      snprintf(value_text, MAX_VALUE_LENGTH, "%d", world->plasma_detail);
      break;
    }

    draw_2d_text(video_texts[index], 5.0, 40.0 + 5.1*(selected - i), 0, index == selected);
    /*
    if (cursor_inside_text())
    {
      manage_video_config_input(world, -(index+1));
      to_change = index;
    }
    */
    if(tf_value != -1)
    {
      if(tf_value)
	draw_2d_text("#V", 70.0, 40 + 5.1*(selected - i), 0, FALSE);
      else
	draw_2d_text("#X", 70.0, 40 + 5.1*(selected - i), 0, FALSE);
    }
    else
      draw_2d_text(value_text, 70.0, 40 + 5.1*(selected - i), 0, FALSE);
  }

  /* UP and DOWN entities */
  draw_2d_text("#U", 5.0, 72.0, 0, FALSE);
  if (cursor_inside_text())
    {
      manage_video_config_input(world, KEY_UP);
    }
  draw_2d_text("#D", 5.0, 8.0, 0, FALSE);
  if (cursor_inside_text())
    {
      manage_video_config_input(world, KEY_DOWN);
    }

  /* mouse selection */
  /*if (to_change >= 0)
    selected = to_change;
  */

  deactivate_2d();
}
